"""Convenience class for standard virtual tiles."""

from iotile.core.exceptions import ArgumentError
from .common_types import tile_rpc
from .virtualtile_base import BaseVirtualTile


class VirtualTile(BaseVirtualTile):
    """A standard virtual tile.

    This convenience class should be the base for most virtual tiles. It
    implements a default status rpc that allows matching proxy objects to the
    tile implementation by name.

    .. important::

        The ``__init__`` signature of all tiles that inherit from this class
        must be: ``__init__(self, address, config, device)`` where ``config``
        is a dictionary of configuration to the tile and device is a
        ``BaseVirtualDevice`` containing the tile.

        This is to allow for TileBasedDevice to uniformly find and create
        these tiles.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
        name (str): The 6 character name that should be returned when this
            tile is asked for its status to allow matching it with a proxy
            object.
        device (BaseVirtualDevice) : optional, device on which this tile is running
    """

    __NO_EXTENSION__ = True

    def __init__(self, address, name, device=None):
        super(VirtualTile, self).__init__(address)
        self.name = _check_convert_name(name)

    @tile_rpc(0x0004, "", "H6sBBBB")
    def tile_status(self):
        """Required status RPC that allows matching a proxy object with a tile."""

        status = (1 << 1) | (1 << 0)  # Configured and running, not currently used but required for compat with physical tiles
        return [0xFFFF, self.name, 1, 0, 0, status]


def _check_convert_name(name):
    if not isinstance(name, bytes):
        name = name.encode('utf-8')
    if len(name) < 6:
        name += b' '*(6 - len(name))
    elif len(name) > 6:
        raise ArgumentError("Virtual tile name is too long, it must be 6 or fewer characters")

    return name



"""A reference implementation of AbstractDeviceServer designed to be subclassed.

:class:`StandardDeviceServer` is a full-featured implementation of the
:class:`AbstractDeviceServer` interface.  See the class docstring for
more information but all that is required to build a concrete device
server on top of StandardDeviceServer is to just handle receiving messages
from clients and calling the appropriate method.  Allocating and freeing
per-client resources is handled internally to make sure that nothing
leaks if clients unexpectedly disconnect.

In particular there is no need to manually keep track of client connections
or manage notifications on behalf of a client.  That is all done internally
inside of StandardDeviceServer.

"""
import logging
import uuid
import inspect
from iotile.core.utilities import SharedLoop
from iotile.core.exceptions import ArgumentError
from ...exceptions import DeviceServerError
from .abstract import AbstractDeviceServer


class StandardDeviceServer(AbstractDeviceServer):
    """A standard reference implementation of AbstractDeviceServer.

    This class is designed to provide the basic set of features that
    any normal AbstractDeviceServer needs.  In particular, it handles
    the three critical aspects of being a device server:

    - managing per-client data and cleaning up after a client is gone
    - managing per-connection data and cleaning up any lingering connections
      after a client is gone.
    - keeping track of which events should be forwarded to which clients.

    It does this by wrapping the underlying methods of the
    AbstractDeviceAdapter that it is given with versions that contain hooks to
    keep track of which client is connected to which device.

    These hooks automatically configure monitors on the underlying adapters
    and take care of removing those monitors when the client connection is
    destroyed for any reason.  Similarly any existing device connections are
    disconnected when a client goes away.

    This standard device server is completely functional on its own with
    no subclassing but developers will typically want to override the
    :meth:`start` and :meth:`stop` routines to actually start and stop their
    particular server (tcp, websockets, mqtt, etc).

    The interface that StandardDeviceServer presents is identical to
    AbstractDeviceAdapter but with the following modifications that make
    it suitable for use in a multi-client server:

    - All methods take a `client_id` parameter so that there is an explicit
      association of all operations with a single client.

    - All methods that previously took a `connection_id` now take a
      `connection_string`.  This is because we cannot guarantee that multiple
      clients would not choose the same connection_id and confuse the device
      adapter.  StandardDeviceServer handles assigning unique connection ids
      to all connections internally.  The connection ids are never exposed
      externally to the clients.  In case multiple device servers are attached
      to the same AbstractDeviceAdapter, they will generate connection ids by
      asking the abstract device adapter for a new unique connection id to
      ensure that they don't collide.

    .. important:

        If you override stop() make sure to call the super() implementation to
        ensure that client resources are properly released when the server
        stops.

    Args:
        adapter (AbstractDeviceAdapter): The underlying device adapter to use
            to find and connect to devices.
        args (dict): Any arguments that you wish to pass to the device server.
        loop (BackgroundEventLoop): The background loop that should be used
            to manage tasks.
    """

    def __init__(self, adapter, args=None, *, loop=SharedLoop):
        super(StandardDeviceServer, self).__init__(adapter, args, loop=loop)

        self.adapter = adapter
        self._clients = {}
        self._loop = loop
        self._logger = logging.getLogger(__name__)

    #pylint:disable=unused-argument;This method is designed to be overridden
    async def client_event_handler(self, client_id, event_tuple, user_data):
        """Method called to actually send an event to a client.

        Users of this class should override this method to actually forward
        device events to their clients.  It is called with the client_id
        passed to (or returned from) :meth:`setup_client` as well as the
        user_data object that was included there.

        The event tuple is a 3-tuple of:

        - connection string
        - event name
        - event object

        If you override this to be acoroutine, it will be awaited.  The
        default implementation just logs the event.

        Args:
            client_id (str): The client_id that this event should be forwarded
                to.
            event_tuple (tuple): The connection_string, event_name and event_object
                that should be forwarded.
            user_data (object): Any user data that was passed to setup_client.
        """

        conn_string, event_name, _event = event_tuple
        self._logger.debug("Ignoring event %s from device %s forwarded for client %s",
                           event_name, conn_string, client_id)

        return None

    def setup_client(self, client_id=None, user_data=None, scan=True, broadcast=False):
        """Setup a newly connected client.

        ``client_id`` must be unique among all connected clients.  If it is
        passed as None, a random client_id will be generated as a string and
        returned.

        This method reserves internal resources for tracking what devices this
        client has connected to and installs a monitor into the adapter on
        behalf of the client.

        It should be called whenever a new client connects to the device server
        before any other activities by that client are allowed.  By default,
        all clients start receiving ``device_seen`` events but if you want
        your client to also receive broadcast events, you can pass broadcast=True.

        Args:
            client_id (str): A unique identifier for this client that will be
                used to refer to it in all future interactions.  If this is
                None, then a random string will be generated for the client_id.
            user_data (object): An arbitrary object that you would like to store
                with this client and will be passed to your event handler when
                events are forwarded to this client.
            scan (bool): Whether to install a monitor to listen for device_found
                events.
            broadcast (bool): Whether to install a monitor to list for broadcast
                events.

        Returns:
            str: The client_id.

            If a client id was passed in, it will be the same as what was passed
            in.  If no client id was passed in then it will be a random unique
            string.
        """

        if client_id is None:
            client_id = str(uuid.uuid4())

        if client_id in self._clients:
            raise  ArgumentError("Duplicate client_id: {}".format(client_id))

        async def _client_callback(conn_string, _, event_name, event):
            event_tuple = (conn_string, event_name, event)
            await self._forward_client_event(client_id, event_tuple)

        client_monitor = self.adapter.register_monitor([], [], _client_callback)

        self._clients[client_id] = dict(user_data=user_data, connections={},
                                        monitor=client_monitor)

        self._adjust_global_events(client_id, scan, broadcast)
        return client_id

    async def start(self):
        """Start the server.

        See :meth:`AbstractDeviceServer.start`.
        """

    async def stop(self):
        """Stop the server and teardown any remaining clients.

        If your subclass overrides this method, make sure to call
        super().stop() to ensure that all devices with open connections from
        thie server are properly closed.

        See :meth:`AbstractDeviceServer.stop`.
        """

        clients = list(self._clients)

        for client in clients:
            self._logger.info("Tearing down client %s at server stop()", client)
            await self.teardown_client(client)

    async def teardown_client(self, client_id):
        """Release all resources held by a client.

        This method must be called and awaited whenever a client is
        disconnected.  It ensures that all of the client's resources are
        properly released and any devices they have connected to are
        disconnected cleanly.

        Args:
            client_id (str): The client that we should tear down.

        Raises:
            ArgumentError: The client_id is unknown.
        """

        client_info = self._client_info(client_id)

        self.adapter.remove_monitor(client_info['monitor'])
        conns = client_info['connections']

        for conn_string, conn_id in conns.items():
            try:
                self._logger.debug("Disconnecting client %s from conn %s at teardown", client_id, conn_string)
                await self.adapter.disconnect(conn_id)
            except:  #pylint:disable=bare-except; This is a finalization method that should not raise unexpectedly
                self._logger.exception("Error disconnecting device during teardown_client: conn_string=%s", conn_string)

        del self._clients[client_id]

    async def probe(self, client_id):
        """Probe for devices on behalf of a client.

        See :meth:`AbstractDeviceAdapter.probe`.

        Args:
            client_id (str): The client we are probing for.

        Raises:
            DeviceServerError: There is an issue with your client_id.
            DeviceAdapterError: The adapter had an issue probing.
        """

        self._client_info(client_id)

        await self.adapter.probe()

    async def connect(self, client_id, conn_string):
        """Connect to a device on behalf of a client.

        See :meth:`AbstractDeviceAdapter.connect`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter to connect.

        Raises:
            DeviceServerError: There is an issue with your client_id.
            DeviceAdapterError: The adapter had an issue connecting.
        """
        conn_id = self.adapter.unique_conn_id()

        self._client_info(client_id)

        await self.adapter.connect(conn_id, conn_string)
        self._hook_connect(conn_string, conn_id, client_id)

    async def disconnect(self, client_id, conn_string):
        """Disconnect from a device on behalf of a client.

        See :meth:`AbstractDeviceAdapter.disconnect`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter to connect.

        Raises:
            DeviceServerError: There is an issue with your client_id such
                as not being connected to the device.
            DeviceAdapterError: The adapter had an issue disconnecting.
        """

        conn_id = self._client_connection(client_id, conn_string)

        try:
            await self.adapter.disconnect(conn_id)
        finally:
            self._hook_disconnect(conn_string, client_id)

    async def open_interface(self, client_id, conn_string, interface):
        """Open a device interface on behalf of a client.

        See :meth:`AbstractDeviceAdapter.open_interface`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter.
            interface (str): The name of the interface to open.

        Raises:
            DeviceServerError: There is an issue with your client_id such
                as not being connected to the device.
            DeviceAdapterError: The adapter had an issue opening the interface.
        """

        conn_id = self._client_connection(client_id, conn_string)

        # Hook first so there is no race on getting the first event
        self._hook_open_interface(conn_string, interface, client_id)
        await self.adapter.open_interface(conn_id, interface)

    async def close_interface(self, client_id, conn_string, interface):
        """Close a device interface on behalf of a client.

        See :meth:`AbstractDeviceAdapter.close_interface`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter.
            interface (str): The name of the interface to close.

        Raises:
            DeviceServerError: There is an issue with your client_id such
                as not being connected to the device.
            DeviceAdapterError: The adapter had an issue closing the interface.
        """

        conn_id = self._client_connection(client_id, conn_string)

        await self.adapter.close_interface(conn_id, interface)
        self._hook_close_interface(conn_string, interface, client_id)

    #pylint:disable=too-many-arguments;This is a legacy method signature.
    async def send_rpc(self, client_id, conn_string, address, rpc_id, payload, timeout):
        """Send an RPC on behalf of a client.

        See :meth:`AbstractDeviceAdapter.send_rpc`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter to connect.
            address (int): The RPC address.
            rpc_id (int): The ID number of the RPC
            payload (bytes): The RPC argument payload
            timeout (float): The RPC's expected timeout to hand to the underlying
                device adapter.

        Returns:
            bytes: The RPC response.

        Raises:
            DeviceServerError: There is an issue with your client_id such
                as not being connected to the device.
            TileNotFoundError: The destination tile address does not exist
            RPCNotFoundError: The rpc_id does not exist on the given tile
            RPCErrorCode: The RPC was invoked successfully and wishes to fail
                with a non-zero status code.
            RPCInvalidIDError: The rpc_id is too large to fit in 16-bits.
            TileBusSerror: The tile was busy and could not respond to the RPC.
            Exception: The rpc raised an exception during processing.
            DeviceAdapterError: If there is a hardware or communication issue
                invoking the RPC.
        """

        conn_id = self._client_connection(client_id, conn_string)
        return await self.adapter.send_rpc(conn_id, address, rpc_id, payload, timeout)

    async def send_script(self, client_id, conn_string, script):
        """Send a script to a device on behalf of a client.

        See :meth:`AbstractDeviceAdapter.send_script`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter.
            script (bytes): The script that we wish to send.

        Raises:
            DeviceServerError: There is an issue with your client_id such
                as not being connected to the device.
            DeviceAdapterError: The adapter had a protocol issue sending the script.
        """

        conn_id = self._client_connection(client_id, conn_string)
        await self.adapter.send_script(conn_id, script)

    async def debug(self, client_id, conn_string, command, args):
        """Send a debug command to a device on behalf of a client.

        See :meth:`AbstractDeviceAdapter.send_script`.

        Args:
            client_id (str): The client we are working for.
            conn_string (str): A connection string that will be
                passed to the underlying device adapter.
            command (str): The name of the debug command to run.
            args (dict): Any command arguments.

        Returns:
            object: The response to the debug command.

        Raises:
            DeviceServerError: There is an issue with your client_id such
                as not being connected to the device.
            DeviceAdapterError: The adapter had a protocol issue sending the debug
                command.
        """

        conn_id = self._client_info(client_id, 'connections')[conn_string]
        return await self.adapter.debug(conn_id, command, args)

    def _adjust_global_events(self, client_id, scan=None, broadcast=None):
        monitor = self._client_info(client_id, 'monitor')

        action = {False: "remove", True: "add"}

        if scan is not None:
            self.adapter.adjust_monitor(monitor, action.get(scan), [None], ['device_seen'])

        if broadcast is not None:
            self.adapter.adjust_monitor(monitor, action.get(broadcast), [None], ['broadcast'])

    #pylint:disable=too-many-arguments;This is an internal utility method
    def _adjust_device_events(self, client_id, conn_string, report=None, trace=None, disconnect=None, progress=None):
        monitor = self._client_info(client_id, 'monitor')

        action = {False: "remove", True: "add"}

        if report is not None:
            self.adapter.adjust_monitor(monitor, action.get(report), [conn_string], ['report'])

        if trace is not None:
            self.adapter.adjust_monitor(monitor, action.get(trace), [conn_string], ['trace'])

        if disconnect is not None:
            self.adapter.adjust_monitor(monitor, action.get(disconnect), [conn_string], ['disconnection'])

        if progress is not None:
            self.adapter.adjust_monitor(monitor, action.get(disconnect), [conn_string], ['progress'])

    def _client_connection(self, client_id, conn_string):
        conns = self._client_info(client_id, 'connections')

        conn_id = conns.get(conn_string)
        if conn_id is None:
            raise DeviceServerError(client_id, conn_string, 'get_connection', 'client does not have connection')

        return conn_id

    def _client_info(self, client_id, key=None):
        client_info = self._clients.get(client_id)

        if client_info is None:
            raise DeviceServerError(client_id, None, 'get_client_info', "unknown client id")

        if key is None:
            return client_info

        return client_info[key]

    def _hook_connect(self, conn_string, conn_id, client_id):
        self._client_info(client_id, 'connections')[conn_string] = conn_id
        self._adjust_device_events(client_id, conn_string, disconnect=True, progress=True)

    def _hook_disconnect(self, conn_string, client_id):
        self._adjust_device_events(client_id, conn_string,
                                   disconnect=False, report=False, trace=False, progress=False)
        del self._client_info(client_id, 'connections')[conn_string]

    def _hook_open_interface(self, conn_string, interface, client_id):
        if interface == 'streaming':
            self._adjust_device_events(client_id, conn_string, report=True)
        elif interface == 'tracing':
            self._adjust_device_events(client_id, conn_string, trace=True)

    def _hook_close_interface(self, conn_string, interface, client_id):
        if interface == 'streaming':
            self._adjust_device_events(client_id, conn_string, report=False)
        elif interface == 'tracing':
            self._adjust_device_events(client_id, conn_string, trace=False)

    async def _forward_client_event(self, client_id, event_tuple):
        conn_string, name, _ = event_tuple
        user_data = self._client_info(client_id, 'user_data')

        if name == 'disconnect':
            self._hook_disconnect(conn_string, client_id)

        if self.client_event_handler is not None:
            result = self.client_event_handler(client_id, event_tuple, user_data=user_data)  #pylint:disable=assignment-from-none;This is an overridable method
            if inspect.isawaitable(result):
                await result



"""Debug and recovery routines that can be used from HardwareManager.

These routines typically only function when you are connected to
and IOTile Device using a special debug adapter, however in that
case they allow memory dumping and forced reprogramming, including
initial code bootstrapping.
"""

import os.path
import os
from enum import IntEnum
import subprocess
import tempfile
import json
from typedargs.annotate import context, docannotate
from iotile.core.exceptions import ArgumentError, ExternalError
from iotile.core.utilities.console import ProgressBar


@context("DebugManager")
class DebugManager:
    """Low level debug operations for development and recovery.

    Args:
        stream (CMDStream): A CMDStream subclass that supports
            debug operations on which enable_debug() has already
            been called.
    """

    def __init__(self, stream):
        self._stream = stream

    def send_command(self, name, cmd_args):
        """Send an arbitrary debug command.

        Args:
            name (str): The name of the debug command to send.
            cmd_args (dict): Any required arguments for the command.
        """

        return self._stream.debug_command(name, cmd_args)

    @docannotate
    def read_memory(self, start_addr, data_length, memory_region="mapped", pause=False):
        """ Read RAM or external memory

        Args:
            memory_region (str): The region of memory to read
            start_addr (integer): The start address to read
            data_length (integer): The length of data to read in bytes
            pause (bool): Optional parameter to halt chip operation
                while performing the core dump.  Defaults to False,
                which could cause rapidly changing RAM values to
                be in an inconsistent state, but it noninvasive and
                will not interrupt any device activity.

        Returns:
            bytes format-as hexdump: The data that was read from the memory region
        """

        # memory = bytearray(self._stream.debug_command(
        #     'read_memory', {'address': start_addr, 'length': data_length}))

        memory_contents = self._stream.debug_command('read_memory', {'region': memory_region, 'start': start_addr, 'length': data_length, 'halt': pause})
        return memory_contents

    @docannotate
    def write_memory(self, start_addr, data, memory_region="mapped"):
        """ Write RAM or external memory in 32bits
        Args:
            memory_region (str): The region of memory to write
            address (integer): The address to write 
            data (bytes): Data to write
        """
        # nbytes = self._stream.debug_command(
        #     'write_memory', {'address': start_addr, 'data': data})
        self._stream.debug_command('write_memory', {'address': start_addr, 'data': data})

    @docannotate
    def dump_memory(self, out_path, start_addr, data_length, memory_region="mapped", pause=False):
        """Dump all RAM or external flash to a binary file.

        Args:
            memory_region (str): The region of memory to dump to file
            out_path (path): The output path at which to save
                the binary core dump.  This core dump consists
                of the current contents of RAM for the device.
            start_addr (integer): The start address to read
            data_length (integer): The length of data to read in bytes
            pause (bool): Optional parameter to halt chip operation
                while performing the core dump.  Defaults to False,
                which could cause rapidly changing RAM values to
                be in an inconsistent state, but it noninvasive and
                will not interrupt any device activity.
        """

        # memory_contents = self._stream.debug_command('dump_memory', {'memory': memory_type, 'start': start_addr, 'length': data_length, 'halt': pause})
        memory_contents = self._stream.debug_command('read_memory', {'region': memory_region, 'start': start_addr, 'length': data_length, 'halt': pause})
        with open(out_path, "wb") as outfile:
            outfile.write(memory_contents)

    @docannotate
    def save_snapshot(self, out_path):
        """Save the current state of an emulated device.

        This debug routine is only supported for emulated devices that
        have a concept of taking a snapshot of their internal state.

        For those devices this will produce a snapshot file that can
        be used in a later call to load_snapshot() in order to reload
        the exact same state.

        Args:
            out_path (path): The output path at which to save
                the binary core dump.  This core dump consists
                of the current contents of RAM for the device.
        """

        internal_state = self.dump_snapshot()
        internal_state = _clean_intenum(internal_state)

        with open(out_path, "w") as outfile:
            json.dump(internal_state, outfile, indent=4)

    @docannotate
    def load_snapshot(self, in_path):
        """Load the current state of an emulated device.

        This debug routine is only supported for emulated devices that
        have a concept of restoring a snapshot of their internal state.

        For those devices this method takes a path to a previously produced
        snapshot file from a call to save_snapshot() and will load that
        snapshot into the currently emulated device.

        Args:
            in_path (path): The output path at which to save
                the binary core dump.  This core dump consists
                of the current contents of RAM for the device.
        """

        with open(in_path, "r") as infile:
            internal_state = json.load(infile)

        self.restore_snapshot(internal_state)

    def dump_snapshot(self):
        """Get the current state of the emulated device.

        This debug routine is only supported for emulated devices that
        have a concept of taking a snapshot of their internal state.

        For those devices this will return a snapshot dictionary with
        the entire internal state of the device.

        Returns:
            dict: The internal snapshot of the device's current state.
        """

        return self._stream.debug_command('dump_state')

    def restore_snapshot(self, snapshot):
        """Restore a previous state snapshot from an emulated device.

        Args:
            snapshot (dict): A snapshot of the internal state of this
                device previously obtained by calling dump_snapshot()
        """

        self._stream.debug_command('restore_state', {'snapshot': snapshot})

    @docannotate
    def open_scenario(self, scenario_path):
        """Load a test scenario from a file into an emulated device.

        This debug routine is only supported for emulated devices and will
        cause the connected device to load one of its preconfigured test
        scenarios.  If the scenario has arguments, the args dict
        may be passed to configure it.

        Args:
            scenario_path (str): The path to a file containing the test
                scenario details.
        """

        with open(scenario_path, "r") as infile:
            scenario = json.load(infile)

        self.load_scenario(scenario)

    def load_scenario(self, scenario_data):
        """Load the given test scenario onto an emulated device.

        This debug routine is only supported for emulated devices and will
        cause the connected device to load one of its preconfigured test
        scenarios.  If the scenario has arguments, the args dict
        may be passed to configure it.

        Args:
            scenario_data (list or dict): Either a dict describing the scenario
                to be loaded or a list of such dicts that will all be loaded.
        """

        self._stream.debug_command('load_scenario', {'scenario': scenario_data})

    @docannotate
    def track_changes(self, enabled=True):
        """Start or stop tracking all internal state changes made to an emulated device.

        This debug routine is only supported for emulated devices and will
        causes the device to create an internal log of all state changes.

        This log can be dumped to a file by calling save_change(output_path).

        Args:
            enabled (bool): Whether we should enable (default) or disable tracking
                changes.
        """

        self._stream.debug_command('track_changes', {'enabled': enabled})

    @docannotate
    def save_changes(self, out_path):
        """Save all tracked changes made to an emulated device.

        This debug routine is only supported for emulated devices and will
        causes the device to save all internal state changes since track_changes
        was called.

        Args:
            out_path (str): The output path where the change log should be saved.
        """

        self._stream.debug_command('dump_changes', {'path': out_path})

    @docannotate
    def flash(self, in_path, file_format=None):
        """Flash a new firmware image to the attached chip.

        This flashing takes place over a debug interface and does not require
        a working bootloader.  In particular, this routine is suitable for
        initial board bring-up of a blank MCU.

        If an explicit format is not passed the following rules are used to
        infer the format based on its file extension.

        elf: ELF file
        hex: intel hex file
        bin: raw binary file assumed to start at address 0

        Args:
            in_path (path): The path to the input firmware image that we wish
                to flash.
            file_format (str): Optional explicit format to use to parse the
                input file.  If this is None, the format is automatically
                inferred from the file suffix.  If given explicitly, you
                should pass 'elf', 'hex' or 'bin'.
        """

        format_map = {
            "elf": self._process_elf,
            "hex": self._process_hex,
            "bin": None
        }

        if file_format is None:
            _root, ext = os.path.splitext(in_path)
            if len(ext) > 0:
                file_format = ext[1:]

        format_handler = format_map.get(file_format)
        if format_handler is None:
            raise ArgumentError("Unknown file format or file extension", file_format=file_format,
                                known_formats=[x for x in format_map if format_map[x] is not None])

        base_addresses, section_data = format_handler(in_path)
        for base_address, data in zip(base_addresses, section_data):
            args = {
                'base_address': base_address,
                'data': data
            }

            progress = ProgressBar("Programming Flash")

            def _progress_callback(finished, total):
                progress.count = total
                progress.progress(finished)

            try:
                progress.start()
                self._stream.debug_command('program_flash', args, _progress_callback)
            finally:
                progress.end()

    @classmethod
    def _process_hex(cls, in_path):
        """This function returns a list of base addresses and a list of the binary data for each segment."""

        from iotile.core.utilities.intelhex import IntelHex

        ihex           = IntelHex(in_path)
        segments       = ihex.segments()
        segments_start = [segment[0] for segment in segments]
        segments_data  = [ihex.tobinarray(start=segment[0], end=segment[1]-1) for segment in segments]

        return segments_start, segments_data

    @classmethod
    def _process_elf(cls, in_path):
        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp.close()

        try:
            err = subprocess.call(['arm-none-eabi-objcopy', '-O', 'ihex', in_path, tmp.name])
            if err != 0:
                raise ExternalError("Cannot convert elf to binary file", error_code=err,
                                    suggestion="Make sure arm-none-eabi-gcc is installed and in your PATH")

            return cls._process_hex(tmp.name)
        finally:
            if os.path.isfile(tmp.name):
                os.remove(tmp.name)


def _clean_intenum(obj):
    """Remove all IntEnum classes from a map."""

    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, IntEnum):
                obj[key] = value.value
            elif isinstance(value, (dict, list)):
                obj[key] = _clean_intenum(value)
    elif isinstance(obj, list):
        for i, value in enumerate(obj):
            if isinstance(value, IntEnum):
                obj[i] = value.value
            elif isinstance(value, (dict, list)):
                obj[i] = _clean_intenum(value)

    return obj



"""Mixin for adding basic event notifications to a DeviceAdapter."""

import inspect
import uuid
import itertools
from iotile.core.exceptions import ArgumentError


class BasicNotificationMixin:
    """A mixin that adds notification support to a DeviceAdapter.

    This mixin class implements the required abstract methods in
    :class:`AbstractDeviceAdapter` related to adding, adjusting and removing
        monitors as well as providing a method :meth:`notify_event` that can
        trigger those monitors.  For situations where you can't await a
        coroutine, there is also a wrapper :meth:`fire_event` that defers the
        notification to a later time inside the BackgroundEventLoop.

    Note that there are not locks inside any of the monitor registraton or
    adjustment procedures so they are only safe to call from within the
    BackgroundEventLoop that is calling the notifications.  This is to ensure
    that the registration/adjustment methods don't need to be coroutines but
    can guarantee to take effect before the next call to ``_notify_callback``.

    .. important::

        There are two requirements on a class that integrates this mixin.

        The class must have a _logger attribute pointing to a logging.Logger
        instance that will be used to log errors.

        It also requires that the class implement a _get_conn_id(conn_string)
        method that can turn a connection_string into a connection id or None
        if there is not an active connection to that device.

        If you also include the standard :class:`PerConnectionDataMixin` mixin
        in your adatper then your adapter will have a compliant implementation
        of ``_get_conn_id()`` and nothing else is needed.

    Args:
        loop (BackgroundEventLoop): The loop we should use to perform
            notifications.
    """

    SUPPORTED_EVENTS = frozenset(['report', 'connection', 'trace', 'disconnection',
                                  'device_seen', 'broadcast', 'progress'])
    SUPPORTED_ADJUSTMENTS = frozenset(['add', 'remove'])
    PROGRESS_OPERATIONS = frozenset(['debug', 'script'])

    def __init__(self, loop):
        self._loop = loop
        self._monitors = {}
        self._callbacks = {}

        self._currently_notifying = False
        self._deferred_adjustments = []

    def register_monitor(self, devices, events, callback):
        """Register a callback when events happen.

        If this method is called, it is guaranteed to take effect before the
        next call to ``_notify_event`` after this method returns.  This method
        is safe to call from within a callback that is itself called by
        ``notify_event``.

        See :meth:`AbstractDeviceAdapter.register_monitor`.
        """

        # Ensure we don't exhaust any iterables
        events = list(events)
        devices = list(devices)

        for event in events:
            if event not in self.SUPPORTED_EVENTS:
                raise ArgumentError("Unknown event type {} specified".format(event), events=events)

        monitor_id = str(uuid.uuid4())

        action = (monitor_id, "add", devices, events)
        self._callbacks[monitor_id] = callback

        if self._currently_notifying:
            self._deferred_adjustments.append(action)
        else:
            self._adjust_monitor_internal(*action)

        return monitor_id

    def _adjust_monitor_internal(self, handle, action, devices, events):
        if action == 'add':
            callback = self._callbacks.get(handle)
            if callback is None:
                self._logger.warning("_adjust_monitor_internal called with an invalid handle, ignoring", handle)

            _add_monitor(self._monitors, handle, callback, devices, events)
        elif action == "remove":
            empty_devices = _remove_monitor(self._monitors, handle, devices, events)

            for device in empty_devices:
                del self._monitors[device]
        elif action == "delete":
            devices, events = _find_monitor(self._monitors, handle)
            self._adjust_monitor_internal(handle, "remove", devices, events)

            if handle in self._callbacks:
                del self._callbacks[handle]
        else:
            pass

    def iter_monitors(self):
        """Iterate over all defined (conn_string, event, monitor) tuples."""

        for conn_string, events in self._monitors.items():
            for event, handlers in events.items():
                for handler in handlers:
                    yield (conn_string, event, handler)

    def adjust_monitor(self, handle, action, devices, events):
        """Adjust a previously registered callback.

        See :meth:`AbstractDeviceAdapter.adjust_monitor`.
        """

        events = list(events)
        devices = list(devices)

        for event in events:
            if event not in self.SUPPORTED_EVENTS:
                raise ArgumentError("Unknown event type {} specified".format(event), events=events)

        if action not in self.SUPPORTED_ADJUSTMENTS:
            raise ArgumentError("Unknown adjustment {} specified".format(action))

        action = (handle, action, devices, events)
        if self._currently_notifying:
            self._deferred_adjustments.append(action)
        else:
            self._adjust_monitor_internal(*action)

    def remove_monitor(self, handle):
        """Remove a previously registered monitor.

        See :meth:`AbstractDeviceAdapter.adjust_monitor`.
        """

        action = (handle, "delete", None, None)
        if self._currently_notifying:
            self._deferred_adjustments.append(action)
        else:
            self._adjust_monitor_internal(*action)

    async def _notify_event_internal(self, conn_string, name, event):
        """Notify that an event has occured.

        This method will send a notification and ensure that all callbacks
        registered for it have completed by the time it returns.  In
        particular, if the callbacks are awaitable, this method will await
        them before returning.  The order in which the callbacks are called
        is undefined.

        This is a low level method that is not intended to be called directly.
        You should use the high level public notify_* methods for each of the
        types of events to ensure consistency in how the event objects are
        created.

        Args:
            conn_string (str): The connection string for the device that the
                event is associated with.
            name (str): The name of the event. Must be in SUPPORTED_EVENTS.
            event (object): The event object.  The type of this object will
                depend on what is being notified.
        """

        try:
            self._currently_notifying = True
            conn_id = self._get_conn_id(conn_string)

            event_maps = self._monitors.get(conn_string, {})
            wildcard_maps = self._monitors.get(None, {})

            wildcard_handlers = wildcard_maps.get(name, {})
            event_handlers = event_maps.get(name, {})

            for handler, func in itertools.chain(event_handlers.items(), wildcard_handlers.items()):
                try:
                    result = func(conn_string, conn_id, name, event)
                    if inspect.isawaitable(result):
                        await result
                except:  #pylint:disable=bare-except;This is a background function and we are logging exceptions
                    self._logger.warning("Error calling notification callback id=%s, func=%s", handler, func, exc_info=True)
        finally:
            for action in self._deferred_adjustments:
                self._adjust_monitor_internal(*action)

            self._deferred_adjustments = []
            self._currently_notifying = False

    def notify_event(self, conn_string, name, event):
        """Notify an event.

        This method will launch a coroutine that runs all callbacks (and
        awaits all coroutines) attached to the given event that was just
        raised.  Internally it uses
        :meth:`BackgroundEventLoop.launch_coroutine` which retains an
        awaitable object when called from within an event loop and a
        concurrent Future object when called outside of the event loop.

        Calling this method from outside of the BackgroundEventLoop is
        considered experimental and not stable behavior that can be depended
        on.

        Args:
            conn_string (str): The connection string for the device that the
                event is associated with.
            name (str): The name of the event. Must be in SUPPORTED_EVENTS.
            event (object): The event object.  The type of this object will
                depend on what is being notified.

        Returns:
            awaitable: An awaitable object that can be used to wait for all callbacks.
        """

        return self._loop.launch_coroutine(self._notify_event_internal, conn_string, name, event)

    def notify_event_nowait(self, conn_string, name, event):
        """Notify an event.

        This will move the notification to the background event loop and
        return immediately.  It is useful for situations where you cannot
        await notify_event but keep in mind that it prevents back-pressure
        when you are notifying too fast so should be used sparingly.

        Note that calling this method will push the notification to a
        background task so it can be difficult to reason about when it will
        precisely occur.  For that reason, :meth:`notify_event` should be
        preferred when possible since that method guarantees that all
        callbacks will be called synchronously before it finishes.

        Args:
            conn_string (str): The connection string for the device that the
                event is associated with.
            name (str): The name of the event. Must be in SUPPORTED_EVENTS.
            event (object): The event object.  The type of this object will
                depend on what is being notified.
        """

        if self._loop.stopping:
            self._logger.debug("Ignoring notification %s from %s because loop is shutting down", name, conn_string)
            return

        self._loop.log_coroutine(self._notify_event_internal, conn_string, name, event)

    #pylint:disable=too-many-arguments;The final wait argument has a sane default
    def notify_progress(self, conn_string, operation, finished, total, wait=True):
        """Send a progress event.

        Progress events can be sent for ``debug`` and ``script`` operations and
        notify the caller about the progress of these potentially long-running
        operations.  They have two integer properties that specify what fraction
        of the operation has been completed.

        Args:
            conn_string (str): The device that is sending the event.
            operations (str): The operation that is in progress: debug or script
            finished (int): The number of "steps" that have finished.
            total (int): The total number of steps to perform.
            wait (bool): Whether to return an awaitable that we can use to
                block until the notification has made it to all callbacks.

        Returns:
            awaitable or None: An awaitable if wait=True.

            If wait is False, the notification is run in the background with
            no way to check its progress and None is returned.
        """

        if operation not in self.PROGRESS_OPERATIONS:
            raise ArgumentError("Invalid operation for progress event: {}".format(operation))

        event = dict(operation=operation, finished=finished, total=total)

        if wait:
            return self.notify_event(conn_string, 'progress', event)

        self.notify_event_nowait(conn_string, 'progress', event)
        return None


def _find_monitor(monitors, handle):
    """Find all devices and events with a given monitor installed."""

    found_devs = set()
    found_events = set()

    for conn_string, device in monitors.items():
        for event, handles in device.items():
            if handle in handles:
                found_events.add(event)
                found_devs.add(conn_string)

    return found_devs, found_events


def _add_monitor(monitors, handle, callback, devices, events):
    """Add the given monitor to the listed devices and events."""

    for conn_string in devices:
        data = monitors.get(conn_string)
        if data is None:
            data = dict()
            monitors[conn_string] = data

        for event in events:
            event_dict = data.get(event)
            if event_dict is None:
                event_dict = dict()
                data[event] = event_dict

            event_dict[handle] = callback


def _remove_monitor(monitors, handle, devices, events):
    """Remove the given monitor from the listed devices and events."""

    empty_devices = []

    for conn_string in devices:
        data = monitors.get(conn_string)
        if data is None:
            continue

        for event in events:
            event_dict = data.get(event)
            if event_dict is None:
                continue

            if handle in event_dict:
                del event_dict[handle]

            if len(event_dict) == 0:
                del data[event]

        if len(data) == 0:
            empty_devices.append(conn_string)

    return empty_devices



"""A utility class for asyncio that lets you await a dict entry."""

import asyncio

_MISSING = object()

class AwaitableDict:
    """Utility class for an async key/value store.

    This class lets you declare a key that should be set at a future point in
    time using declare(name).  Then you can wait until the key is set with an
    optional maximum timeout.  Another task or callback in the same loop can
    set the value using set() and your blocking get() call will automatically
    return the value.

    This class is useful for blocking on long-running network operations that
    have a unique identifier associated with them.  The method are optimized
    to be easy to use.  With the default keyword parameters, there is no need
    to clean anything up manually.

    Args:
        loop (asyncio.BackgroundEventLoop): The loop that this object should
            be associated with.
    """

    def __init__(self, loop):
        self._data = {}
        self._loop = loop

    def __len__(self):
        return len(self._data)

    def _ensure_declared(self, name):
        if name in self._data:
            return

        self.declare(name)

    def declare(self, name):
        """Declare that a key will be set in the future.

        This will create a future for the key that is used to
        hold its result and allow awaiting it.

        Args:
            name (str): The unique key that will be used.
        """

        if name in self._data:
            raise KeyError("Declared name {} that already existed".format(name))

        self._data[name] = self._loop.create_future()

    async def get(self, name, timeout=None, autoremove=True):
        """Wait for a value to be set for a key.

        This is the primary way to receive values from AwaitableDict.
        You pass in the name of the key you want to wait for, the maximum
        amount of time you want to wait and then you can await the result
        and it will resolve to value from the call to set or an
        asyncio.TimeoutError.

        You should generally leave autoremove as the default True value. This
        causes the key to be removed from the dictionary after get returns.
        Normally you have a single user calling ``get`` and another calling
        ``set`` so you want to automatically clean up after the getter
        returns, no matter what.

        If the key has not already been declared, it will be declared
        automatically inside this function so it is not necessary to call
        :meth:`declare` manually in most use cases.

        Args:
            name (str): The name of the key to wait on.
            timeout (float): The maximum timeout to wait.
            autoremove (bool): Whether to automatically remove the
                key when get() returns.

        Returns:
            object: Whatever was set in the key by :meth:`set`.

        Raises:
            asyncio.TimeoutError: The key was not set within the timeout.
        """

        self._ensure_declared(name)

        try:
            await asyncio.wait_for(self._data[name], timeout, loop=self._loop.get_loop())
            return self._data[name].result()
        finally:
            if autoremove:
                self._data[name].cancel()
                del self._data[name]

    def get_nowait(self, name, default=_MISSING, autoremove=False):
        """Get the value of a key if it is already set.

        This method allows you to check if a key has already been set
        without blocking.  If the key has not been set you will get the
        default value you pass in or KeyError() if no default is passed.

        When this method returns the key is automatically removed unless
        you pass ``autoremove=False``.

        This method is not a coroutine and does not block.

        Args:
            name (str): The name of the key to wait on.
            default (object): The default value to return if the key
                has not yet been set.  Defaults to raising KeyError().
            autoremove (bool): Whether to automatically remove the
                key when get() returns.

        Returns:
            object: Whatever was set in the key by :meth:`set`.
        """

        self._ensure_declared(name)

        try:
            future = self._data[name]
            if future.done():
                return future.result()

            if default is _MISSING:
                raise KeyError("Key {} has not been assigned a value and no default given".format(name))

            return default
        finally:
            if autoremove:
                self._data[name].cancel()
                del self._data[name]

    def set(self, name, value, autodeclare=False):
        """Set the value of a key.

        This method will cause anyone waiting on a key (and any future
        waiters) to unblock and be returned the value you pass here.

        If the key has not been declared previously, a KeyError() is
        raised unless you pass ``autodeclare=True`` which will cause
        the key to be declared.  Normally you don't want to autodeclare.

        This method is not a coroutine and does not block.

        Args:
            name (str): The key to set
            value (object): The value to set
            autodeclare (bool): Whether to automatically declare the
                key if is has not already been declared.  Defaults to
                False.
        """

        if not autodeclare and name not in self._data:
            raise KeyError("Key {} has not been declared and autodeclare=False".format(name))

        self._ensure_declared(name)
        self._data[name].set_result(value)



"""A global background asyncio event loop for CoreTools.

This class is the base primitive used by all of CoreTools to interact with the
asyncio framework.  There should be a single SharedLoop per process using
CoreTools in an asynchronous fashion.

Various subsystems add tasks to this background event loop which runs until
the process shuts down or until you explicitly stop it using SharedLoop.stop().

The primary mode of interaction with the event loop is by scheduling long
running tasks using SharedLoop.add_task().  You get a BackgroundTask object
back that you can stop at any time from any thread using BackgroundTask.stop().

If you have complex tasks that must be stopped in a specific order, you can
register then as subtasks and their parent task is in charge of stopping
them cleanly in the correct order.
"""

import time
import asyncio
import inspect
import logging
import threading
import atexit
import functools
import concurrent.futures
from iotile.core.exceptions import TimeoutExpiredError, ArgumentError, InternalError, LoopStoppingError


class BackgroundTask:
    """A background coroutine task running in a BackgroundEventLoop.

    This class is a wrapper around asyncio.Task that allows tasks to have
    subtasks.  A subtask is just a regular task but the parent task is in
    charge of stopping it when required.  This distinction allows you to
    create hierarchies of tasks that can be cleanly stopped in a known
    order when the BackgroundEventLoop is stopped.

    Tasks can also be given names that will be logged for debugging purposes.

    This class should never be created directly but will be returned
    by BackgroundEventLoop.add_task().

    Generally, you always want to start a task with a coroutine that it
    runs.  However, there are cases where you just want a "parent" task
    as a placeholder that can group various subtasks that do have coroutines.

    In that case, you can pass ``cor`` as None to say that there is no
    underlying asnycio task backing this task.  In that case, you **should**
    pass ``finalizer`` as something other than None.  You can create a task
    that has no coroutine and no finalizer, which is basically just a
    placeholder that does nothing.

    Args:
        cor (coroutine or asyncio.Task): An asyncio Task or the coroutine
            that we should execute as a task.  If a coroutine is given
            it is scheduled as a task in threadsafe manner automatically.
        name (str): The name of the task for pretty printing and debug
            purposes.  If not specified, it defaults to the underlying
            asyncio task object instance name.
        finalizer (callable): An optional callable that should be
            invoked to cancel the task.  If specified it will be called
            with a single argument, this BackgroundTask instance.

            If not specified, calling stop() will result in cancel()
            being called on the underlying task.

            The finalizer (or task.cancel()) is always invoked inside
            the event loop.  If finalizer is a coroutine it is awaited.
        loop (BackgroundEventLoop): The background event loop this task
            should run in.  If not specified, it defaults to the global
            SharedLoop instance.
        stop_timeout (float): The maximum amount of time to wait for this
            task to stop when stop() is called in seconds.  None indicates
            an unlimited amount of time.  Default is 1 second.
    """

    #pylint:disable=too-many-arguments;This class is not meant to be directly constructed by the user
    def __init__(self, cor, name=None, finalizer=None, stop_timeout=1.0, loop=None):
        self._name = name
        self._finalizer = finalizer
        self._stop_timeout = stop_timeout
        self._logger = logging.getLogger(__name__)
        self.stopped = False

        if loop is None:
            loop = SharedLoop

        if not isinstance(loop, BackgroundEventLoop):
            raise ArgumentError("A BackgroundTask must be created with a BackgroundEventLoop, loop={}".format(loop))

        self._loop = loop

        self.subtasks = []
        if inspect.iscoroutine(cor):
            self.task = _create_task_threadsafe(cor, self._loop)
        elif inspect.iscoroutinefunction(cor):
            self.task = _create_task_threadsafe(cor(), self._loop)
        elif isinstance(cor, asyncio.Task):
            self.task = cor
        elif cor is None:
            self.task = None
        else:
            raise ArgumentError("Unknown object passed to Background task: {}".format(cor))

    @property
    def name(self):
        """A descriptive name for this task."""

        if self._name is not None:
            return self._name

        return str(self.task)

    def create_subtask(self, cor, name=None, stop_timeout=1.0):
        """Create and add a subtask from a coroutine.

        This function will create a BackgroundTask and then
        call self.add_subtask() on it.

        Args:
            cor (coroutine): The coroutine that should be wrapped
                in a background task.
            name (str): An optional name for the task.
            stop_timeout (float): The maximum time to wait for this
                subtask to die after stopping it.

        Returns:
            Backgroundtask: The created subtask.
        """

        if self.stopped:
            raise InternalError("Cannot add a subtask to a parent that is already stopped")

        subtask = BackgroundTask(cor, name, loop=self._loop, stop_timeout=stop_timeout)
        self.add_subtask(subtask)
        return subtask

    def add_subtask(self, subtask):
        """Link a subtask to this parent task.

        This will cause stop() to block until the subtask has also
        finished.  Calling stop will not directly cancel the subtask.
        It is expected that your finalizer for this parent task will
        cancel or otherwise stop the subtask.

        Args:
            subtask (BackgroundTask): Another task that will be stopped
                when this task is stopped.
        """

        if self.stopped:
            raise InternalError("Cannot add a subtask to a parent that is already stopped")

        if not isinstance(subtask, BackgroundTask):
            raise ArgumentError("Subtasks must inherit from BackgroundTask, task={}".format(subtask))

        #pylint:disable=protected-access;It is the same class as us so is equivalent to self access.
        if subtask._loop != self._loop:
            raise ArgumentError("Subtasks must run in the same BackgroundEventLoop as their parent",
                                subtask=subtask, parent=self)

        self.subtasks.append(subtask)

    async def stop(self):
        """Stop this task and wait until it and all its subtasks end.

        This function will finalize this task either by using the finalizer
        function passed during creation or by calling task.cancel() if no
        finalizer was passed.

        It will then call join() on this task and any registered subtasks
        with the given maximum timeout, raising asyncio.TimeoutError if
        the tasks did not exit within the given timeout.

        This method should only be called once.

        After this method returns, the task is finished and no more subtasks
        can be added.  If this task is being tracked inside of the
        BackgroundEventLoop that it is part of, it will automatically be
        removed from the event loop's list of tasks.
        """

        if self.stopped:
            return

        self._logger.debug("Stopping task %s", self.name)

        if self._finalizer is not None:
            try:
                result = self._finalizer(self)
                if inspect.isawaitable(result):
                    await result
            except:  #pylint:disable=bare-except;We need to make sure we always wait for the task
                self._logger.exception("Error running finalizer for task %s",
                                       self.name)
        elif self.task is not None:
            self.task.cancel()

        tasks = []
        if self.task is not None:
            tasks.append(self.task)

        tasks.extend(x.task for x in self.subtasks)
        finished = asyncio.gather(*tasks, return_exceptions=True)
        outcomes = []

        try:
            outcomes = await asyncio.wait_for(finished, timeout=self._stop_timeout)
        except asyncio.TimeoutError as err:
            # See discussion here: https://github.com/python/asyncio/issues/253#issuecomment-120138132
            # This prevents a nuisance log error message, finished is guaranteed
            # to be cancelled but not awaited when wait_for() has a timeout.
            try:
                outcomes = await finished
            except asyncio.CancelledError:
                pass

            # See https://mail.python.org/pipermail/python-3000/2008-May/013740.html
            # for why we need to explictly name the error here
            raise err
        finally:
            self.stopped = True
            for outcome in outcomes:
                if isinstance(outcome, Exception) and not isinstance(outcome, asyncio.CancelledError):
                    self._logger.error(outcome)

            if self in self._loop.tasks:
                self._loop.tasks.remove(self)

    def stop_threadsafe(self):
        """Stop this task from another thread and wait for it to finish.

        This method must not be called from within the BackgroundEventLoop but
        will inject self.stop() into the event loop and block until it
        returns.

        Raises:
            TimeoutExpiredError: If the task does not stop in the given
                timeout specified in __init__()
        """

        if self.stopped:
            return

        try:
            self._loop.run_coroutine(self.stop())
        except asyncio.TimeoutError:
            raise TimeoutExpiredError("Timeout stopping task {} with {} subtasks".format(self.name, len(self.subtasks)))


class BackgroundEventLoop:
    """A shared asyncio event loop running in a background thread.

    This class represents a single background event loop.

    The background thread is created the first time a request is made that
    requires a loop to present so it is cheap to create a BackgroundEventLoop
    and not use it in any way.

    There is also a global instance of BackgroundEventLoop called SharedLoop
    that can be used when you just want to add tasks to a single shared loop.
    Creating your own BackgroundEventLoop should not be generally done unless
    you are unit testing.

    A background event loop cannot be restarted once it stops.  Once the shutdown
    process is started by calling ``stop()``, no more tasks can be added to the
    loop and attempts to add them will raise an InternalError.

    This prevents people from accidentally adding more tasks while the loop is
    shutting down.
    """

    def __init__(self):
        self.loop = None
        self.thread = None
        self.stopping = False
        self.tasks = set()

        self._logger = logging.getLogger(__name__)
        self._loop_check = threading.local()
        self._pool = None

    def start(self, aug='EventLoopThread'):
        """Ensure the background loop is running.

        This method is safe to call multiple times.  If the loop is already
        running, it will not do anything.
        """

        if self.stopping:
            raise LoopStoppingError("Cannot perform action while loop is stopping.")

        if not self.loop:
            self._logger.debug("Starting event loop")
            self.loop = asyncio.new_event_loop()
            self.thread = threading.Thread(target=self._loop_thread_main, name=aug, daemon=True)
            self.thread.start()

    def wait_for_interrupt(self, check_interval=1.0, max_time=None):
        """Run the event loop until we receive a ctrl-c interrupt or max_time passes.

        This method will wake up every 1 second by default to check for any
        interrupt signals or if the maximum runtime has expired.  This can be
        set lower for testing purpose to reduce latency but in production
        settings, this can cause increased CPU usage so 1 second is an
        appropriate value.

        Args:
            check_interval (float): How often to wake up and check for
                a SIGTERM. Defaults to 1s.  Setting this faster is useful
                for unit testing.  Cannot be < 0.01 s.
            max_time (float): Stop the event loop after max_time seconds.
                This is useful for testing purposes.  Defaults to None,
                which means run forever until interrupt.
        """

        self.start()

        wait = max(check_interval, 0.01)
        accum = 0

        try:
            while max_time is None or accum < max_time:
                try:
                    time.sleep(wait)
                except IOError:
                    pass  # IOError comes when this call is interrupted in a signal handler

                accum += wait
        except KeyboardInterrupt:
            pass

    def stop(self):
        """Synchronously stop the background loop from outside.

        This method will block until the background loop is completely stopped
        so it cannot be called from inside the loop itself.

        This method is safe to call multiple times.  If the loop is not
        currently running it will return without doing anything.
        """

        if not self.loop:
            return

        if self.inside_loop():
            raise InternalError("BackgroundEventLoop.stop() called from inside event loop; "
                                "would have deadlocked.")

        try:
            self.run_coroutine(self._stop_internal())
            self.thread.join()

            if self._pool is not None:
                self._pool.shutdown(wait=True)
        except:
            self._logger.exception("Error stopping BackgroundEventLoop")
            raise
        finally:
            self.thread = None
            self.loop = None
            self._pool = None
            self.tasks = set()

    def get_loop(self):
        """Get the current loop instance.

        If there is no current loop, a new loop is created and started before
        returning.

        Returns:
            asyncio.Loop: The loop.
        """

        if not self.loop:
            self.start()

        return self.loop

    def inside_loop(self):
        """Check if we are running inside the event loop.

        This can be used to decide whether we need to use a threadsafe method
        to interact with the loop or whether we can just interact with it
        directly. It is used inside BackgroundEventLoop to check for situations that
        would deadlock and raise an Exception instead.

        Returns:
            bool: True if we are running inside the loop thread.
        """

        return self._loop_check.__dict__.get('inside_loop', False)

    async def _stop_internal(self):
        """Cleanly stop the event loop after shutting down all tasks."""

        # Make sure we only try to stop once
        if self.stopping is True:
            return

        self.stopping = True

        awaitables = [task.stop() for task in self.tasks]
        results = await asyncio.gather(*awaitables, return_exceptions=True)
        for task, result in zip(self.tasks, results):
            if isinstance(result, Exception):
                self._logger.error("Error stopping task %s: %s", task.name, repr(result))

        # It is important to defer this call by one loop cycle so
        # that this coroutine is finalized and anyone blocking on it
        # resumes execution.
        self.loop.call_soon(self.loop.stop)

    def _loop_thread_main(self):
        """Main background thread running the event loop."""

        asyncio.set_event_loop(self.loop)
        self._loop_check.inside_loop = True

        try:
            self._logger.debug("Starting loop in background thread")
            self.loop.run_forever()
            self._logger.debug("Finished loop in background thread")
        except:  # pylint:disable=bare-except;This is a background worker thread.
            self._logger.exception("Exception raised from event loop thread")
        finally:
            self.loop.close()

    # pylint:disable=too-many-arguments;These all have sane defaults that should not be changed often.
    def add_task(self, cor, name=None, finalizer=None, stop_timeout=1.0, parent=None):
        """Schedule a task to run on the background event loop.

        This method will start the given coroutine as a task and  keep track
        of it so that it can be properly shutdown which the event loop is
        stopped.

        If parent is None, the task will be stopped by calling finalizer()
        inside the event loop and then awaiting the task.  If finalizer is
        None then task.cancel() will be called to stop the task.  If finalizer
        is specified, it is called with a single argument (self, this
        BackgroundTask).  Finalizer can be a simple function, or any
        awaitable.  If it is an awaitable it will be awaited.

        If parent is not None, it must be a BackgroundTask object previously
        created by a call to BackgroundEventLoop.add_task() and this task will be
        registered as a subtask of that task.  It is that task's job then to
        cancel this task or otherwise stop it when it is stopped.

        This method is safe to call either from inside the event loop itself
        or from any other thread without fear of deadlock or race.

        Args:
            cor (coroutine or asyncio.Task): An asyncio Task or the coroutine
                that we should execute as a task.  If a coroutine is given
                it is scheduled as a task in threadsafe manner automatically.
            name (str): The name of the task for pretty printing and debug
                purposes.  If not specified, it defaults to the underlying
                asyncio task object instance name.
            finalizer (callable): An optional callable that should be
                invoked to cancel the task.  If not specified, calling stop()
                will result in cancel() being called on the underlying task.


            stop_timeout (float): The maximum amount of time to wait for this
                task to stop when stop() is called in seconds.  None indicates
                an unlimited amount of time.  Default is 1.

                This is ignored if parent is not None.
            parent (BackgroundTask): A previously created task that will take
                responsibility for stopping this task when it is stopped.

        Returns:
            BackgroundTask: The BackgroundTask representing this task.
        """

        if self.stopping:
            raise LoopStoppingError("Cannot add task because loop is stopping")

        # Ensure the loop exists and is started
        self.start()

        if parent is not None and parent not in self.tasks:
            raise ArgumentError("Designated parent task {} is not registered".format(parent))

        task = BackgroundTask(cor, name, finalizer, stop_timeout, loop=self)
        if parent is None:
            self.tasks.add(task)
            self._logger.debug("Added primary task %s", task.name)
        else:
            parent.add_subtask(task)
            self._logger.debug("Added subtask %s to parent %s", task.name, parent.name)

        return task

    def run_coroutine(self, cor, *args, **kwargs):
        """Run a coroutine to completion and return its result.

        This method may only be called outside of the event loop.
        Attempting to call it from inside the event loop would deadlock
        and will raise InternalError instead.

        Args:
            cor (coroutine): The coroutine that we wish to run in the
                background and wait until it finishes.

        Returns:
            object: Whatever the coroutine cor returns.
        """

        if self.stopping:
            raise LoopStoppingError("Could not launch coroutine because loop is shutting down: %s" % cor)

        self.start()

        cor = _instaniate_coroutine(cor, args, kwargs)

        if self.inside_loop():
            raise InternalError("BackgroundEventLoop.run_coroutine called from inside event loop, "
                                "would have deadlocked.")

        future = self.launch_coroutine(cor)
        return future.result()

    def launch_coroutine(self, cor, *args, **kwargs):
        """Start a coroutine task and return a blockable/awaitable object.

        If this method is called from inside the event loop, it will return an
        awaitable object.  If it is called from outside the event loop it will
        return an concurrent Future object that can block the calling thread
        until the operation is finished.

        Args:
            cor (coroutine): The coroutine that we wish to run in the
                background and wait until it finishes.

        Returns:
            Future or asyncio.Task: A future representing the coroutine.

            If this method is called from within the background loop
            then an awaitable asyncio.Tasks is returned.  Otherwise,
            a concurrent Future object is returned that you can call
            ``result()`` on to block the calling thread.
        """

        if self.stopping:
            raise LoopStoppingError("Could not launch coroutine because loop is shutting down: %s" % cor)

        # Ensure the loop exists and is started
        self.start()

        cor = _instaniate_coroutine(cor, args, kwargs)

        if self.inside_loop():
            return asyncio.ensure_future(cor, loop=self.loop)

        return asyncio.run_coroutine_threadsafe(cor, loop=self.loop)

    def run_in_executor(self, func, *args, **kwargs):
        """Execute a function on a background executor.

        This method is used to run blocking functions in an awaitable fashion.
        The function is dispatched to a background worker thread and an
        awaitable is returned that will finalize when the function has
        finished.
        """

        self.start()

        if self._pool is None:
            self._pool = concurrent.futures.ThreadPoolExecutor()

        return self.loop.run_in_executor(self._pool, functools.partial(func, *args, **kwargs))

    def log_coroutine(self, cor, *args, **kwargs):
        """Run a coroutine logging any exception raised.

        This routine will not block until the coroutine is finished
        nor will it return any result.  It will just log if any
        exception is raised by the coroutine during operation.

        It is safe to call from both inside and outside the event loop.

        There is no guarantee on how soon the coroutine will be scheduled.

        Args:
            cor (coroutine): The coroutine that we wish to run in the
                background and wait until it finishes.
        """

        if self.stopping:
            raise LoopStoppingError("Could not launch coroutine because loop is shutting down: %s" % cor)

        self.start()

        cor = _instaniate_coroutine(cor, args, kwargs)

        def _run_and_log():
            task = self.loop.create_task(cor)
            task.add_done_callback(lambda x: _log_future_exception(x, self._logger))

        if self.inside_loop():
            _run_and_log()
        else:
            self.loop.call_soon_threadsafe(_run_and_log)

    def create_event(self):
        """Attach an Event to the background loop.

        Returns:
            asyncio.Event
        """

        self.start()
        return asyncio.Event(loop=self.loop)

    def create_lock(self):
        """Attach a Lock to the background loop.

        Returns:
            asyncio.Lock
        """

        self.start()
        return asyncio.Lock(loop=self.loop)

    def create_future(self):
        """Attach a Future to the background loop.

        Returns:
            asyncio.Future
        """

        self.start()
        return asyncio.Future(loop=self.loop)

    def create_queue(self):
        """Attach a Queue to the background loop.

        Returns:
            asyncio.Queue
        """

        self.start()
        return asyncio.Queue(loop=self.loop)


def _create_task_threadsafe(cor, loop):
    asyncio_loop = loop.get_loop()

    if loop.inside_loop():
        return asyncio_loop.create_task(cor)

    async def _task_creator():
        return asyncio_loop.create_task(cor)

    future = asyncio.run_coroutine_threadsafe(_task_creator(), loop=asyncio_loop)
    return future.result()


def _instaniate_coroutine(cor, args, kwargs):
    if inspect.iscoroutinefunction(cor):
        cor = cor(*args, **kwargs)
    elif len(args) > 0 or len(kwargs) > 0:
        raise ArgumentError("You cannot pass arguments if coroutine is already created", args=args, kwargs=kwargs)

    return cor


def _log_future_exception(future, logger):
    """Log any exception raised by future."""

    if not future.done():
        return

    try:
        future.result()
    except:  #pylint:disable=bare-except;This is a background logging helper
        logger.warning("Exception in ignored future: %s", future, exc_info=True)


# Create a single global event loop that anyone can add tasks to.
SharedLoop = BackgroundEventLoop()  # pylint:disable=invalid-name;This is for backwards compatibility.

# Ensure that the concurrent atexit handler is registered before our atexit handler.
#
# If we register a task that uses the concurrent API to run things on executor
# threads, those threads need to be still running when we stop the event loop
# atexit in case they are used during the shutdown process.
#
# The order of atexit handlers is the opposite of the order in which they are
# registered, so ensure the thread handler gets registered first.
#
# See: https://github.com/python/cpython/blob/master/Lib/concurrent/futures/thread.py#L33

# pylint:disable=wrong-import-position,wrong-import-order,unused-import;See above comment
import concurrent.futures.thread
atexit.register(SharedLoop.stop)



"""A wrapper class that turns an AbstractDeviceAdapter into a legacy DeviceAdapter.

This allows modern AbstractDeviceAdapter classes to be used inside
AdapterStream so that they can be used by the synchronous HardwareManager.
"""

import asyncio
import logging
from iotile.core.exceptions import ArgumentError
from iotile.core.utilities import SharedLoop

from .abstract import AbstractDeviceAdapter
from .legacy import DeviceAdapter
from ...exceptions import (DeviceAdapterError, RPCInvalidIDError,
                           TileNotFoundError, RPCNotFoundError,
                           RPCErrorCode, BusyRPCResponse)
from ...virtual import pack_rpc_response

_MISSING = object()


class SynchronousLegacyWrapper(DeviceAdapter):
    """Provide a legacy synchronous API on top of an AbstractDeviceAdapter."""

    MAX_ADAPTER_STARTUP_TIME = 10.0

    def __init__(self, adapter: AbstractDeviceAdapter, loop=SharedLoop):
        super(SynchronousLegacyWrapper, self).__init__()

        if not isinstance(adapter, AbstractDeviceAdapter):
            raise ArgumentError("DeviceAdapterLegacyWrapper created from object "
                                "not a subclass of AbstractDeviceAdapter", adapter=adapter)

        self._adapter = adapter
        self._loop = loop
        self._logger = logging.getLogger(__name__)

        # Synchronously start the adapter and block until it is started
        future = self._loop.launch_coroutine(self._adapter.start())
        future.result(SynchronousLegacyWrapper.MAX_ADAPTER_STARTUP_TIME)

    def set_config(self, name, value):
        """Set a config value for this adapter by name

        Args:
            name (string): The name of the config variable
            value (object): The value of the config variable
        """

        self._adapter.set_config(name, value)

    def get_config(self, name, default=_MISSING):
        """Get a config value from this adapter by name

        Args:
            name (string): The name of the config variable
            default (object): The default value to return if config is not found

        Returns:
            object: the value associated with the name

        Raises:
            ArgumentError: if the name is not found and no default is supplied
        """

        value = self._adapter.get_config(name, default)
        if value is _MISSING:
            raise ArgumentError("Config value did not exist", name=name)

        return value

    def add_callback(self, name, func):
        """Add a callback when device events happen.

        Args:
            name (str): currently support 'on_scan' and 'on_disconnect'
            func (callable): the function that should be called
        """

        if name == 'on_scan':
            events = ['device_seen']
            def callback(_conn_string, _conn_id, _name, event):
                func(self.id, event, event.get('validity_period', 60))
        elif name == 'on_report':
            events = ['report', 'broadcast']
            def callback(_conn_string, conn_id, _name, event):
                func(conn_id, event)
        elif name == 'on_trace':
            events = ['trace']
            def callback(_conn_string, conn_id, _name, event):
                func(conn_id, event)
        elif name == 'on_disconnect':
            events = ['disconnection']
            def callback(_conn_string, conn_id, _name, _event):
                func(self.id, conn_id)
        else:
            raise ArgumentError("Unknown callback type {}".format(name))

        self._adapter.register_monitor([None], events, callback)

    def can_connect(self):
        return self._adapter.can_connect()

    def periodic_callback(self):
        """Periodic callback to allow adapter to process events.

        This is a no-op on asynchronous device adapters since they are able to
        run their own periodic callbacks inside the background event loop.
        """

    def stop_sync(self):
        future = self._loop.launch_coroutine(self._adapter.stop())
        return self._format_future(future)

    def connect_async(self, conn_id, connection_string, callback):
        """Asynchronously connect to a device."""

        future = self._loop.launch_coroutine(self._adapter.connect(conn_id, connection_string))
        future.add_done_callback(lambda x: self._callback_future(conn_id, x, callback))

    def disconnect_async(self, conn_id, callback):
        """Asynchronously disconnect from a device."""

        future = self._loop.launch_coroutine(self._adapter.disconnect(conn_id))
        future.add_done_callback(lambda x: self._callback_future(conn_id, x, callback))

    def open_interface_async(self, conn_id, interface, callback, connection_string=None):
        """Asynchronously connect to a device."""

        future = self._loop.launch_coroutine(self._adapter.open_interface(conn_id, interface))
        future.add_done_callback(lambda x: self._callback_future(conn_id, x, callback))

    def probe_async(self, callback):
        """Asynchronously connect to a device."""

        future = self._loop.launch_coroutine(self._adapter.probe())
        future.add_done_callback(lambda x: self._callback_future(None, x, callback))

    def send_rpc_async(self, conn_id, address, rpc_id, payload, timeout, callback):
        """Asynchronously send an RPC to this IOTile device."""

        future = self._loop.launch_coroutine(self._adapter.send_rpc(conn_id, address, rpc_id, payload, timeout))

        def format_response(future):
            payload = None
            exception = future.exception()
            rpc_status = None
            rpc_response = b''
            failure = None
            success = True

            if exception is None:
                payload = future.result()
                rpc_status, rpc_response = pack_rpc_response(payload, exception)
            elif isinstance(exception, (RPCInvalidIDError, TileNotFoundError, RPCNotFoundError,
                                        RPCErrorCode, BusyRPCResponse)):
                rpc_status, rpc_response = pack_rpc_response(payload, exception)
            else:
                success = False
                failure = str(exception)

            callback(conn_id, self.id, success, failure, rpc_status, rpc_response)

        future.add_done_callback(format_response)

    def debug_async(self, conn_id, cmd_name, cmd_args, progress_callback, callback):
        """Asynchronously complete a named debug command.

        The command name and arguments are passed to the underlying device adapter
        and interpreted there.  If the command is long running, progress_callback
        may be used to provide status updates.  Callback is called when the command
        has finished.

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            cmd_name (string): the name of the debug command we want to invoke
            cmd_args (dict): any arguments that we want to send with this command.
            progress_callback (callable): A function to be called with status on our progress, called as:
                progress_callback(done_count, total_count)
            callback (callable): The callback that should be called when finished.
        """

        def monitor_callback(_conn_string, _conn_id, _event_name, event):
            if event.get('operation') != 'debug':
                return

            progress_callback(event.get('finished'), event.get('total'))

        async def _install_monitor():
            try:
                conn_string = self._adapter._get_property(conn_id, 'connection_string')
                return self._adapter.register_monitor([conn_string], ['progress'], monitor_callback)
            except:  #pylint:disable=bare-except;This is a legacy shim that must always ensure it doesn't raise.
                self._logger.exception("Error installing debug progress monitor")
                return None

        monitor_id = self._loop.run_coroutine(_install_monitor())
        if monitor_id is None:
            callback(conn_id, self.id, False, 'could not install progress monitor', None)
            return

        future = self._loop.launch_coroutine(self._adapter.debug(conn_id, cmd_name, cmd_args))

        def format_response(future):
            ret_val = None
            success = True
            failure = None
            if future.exception() is None:
                ret_val = future.result()
            else:
                success = False
                failure = str(future.exception())

            self._adapter.remove_monitor(monitor_id)
            callback(conn_id, self.id, success, ret_val, failure)

        future.add_done_callback(format_response)

    def send_script_async(self, conn_id, data, progress_callback, callback):
        """Asynchronously send a script to the device."""

        def monitor_callback(_conn_string, _conn_id, _event_name, event):
            if event.get('operation') != 'script':
                return

            progress_callback(event.get('finished'), event.get('total'))

        async def _install_monitor():
            try:
                conn_string = self._adapter._get_property(conn_id, 'connection_string')
                return self._adapter.register_monitor([conn_string], ['progress'], monitor_callback)
            except:  #pylint:disable=bare-except;This is a legacy shim that must always ensure it doesn't raise.
                self._logger.exception("Error installing script progress monitor")
                return None

        monitor_id = self._loop.run_coroutine(_install_monitor())
        if monitor_id is None:
            callback(conn_id, self.id, False, 'could not install progress monitor')
            return

        future = self._loop.launch_coroutine(self._adapter.send_script(conn_id, data))
        future.add_done_callback(lambda x: self._callback_future(conn_id, x, callback, monitors=[monitor_id]))

    def _format_exception(self, exception):
        if isinstance(exception, DeviceAdapterError):
            reason = exception.reason
        elif isinstance(exception, asyncio.TimeoutError):
            reason = "operation timed out"
        else:
            self._logger.error("Unknown exception during AbstractDeviceAdapter operation: %s", exception)
            reason = "Unknown exception {}".format(exception)

        return {
            'success': False,
            'failure_reason': reason
        }

    def _format_future(self, future):
        exception = future.exception()
        if exception is not None:
            return self._format_exception(exception)

        return {
            'success': True,
            'failure_reason': None
        }

    def _callback_future(self, conn_id, future, callback, monitors=None):
        info = self._format_future(future)

        if monitors is not None:
            for monitor in monitors:
                self._adapter.remove_monitor(monitor)

        if conn_id is not None:
            callback(conn_id, self.id, info.get('success'), info.get('failure_reason'))
        else:
            callback(self.id, info.get('success'), info.get('failure_reason'))



"""Helper module for assigning UTC times to readings without a realtime clock.

Some IOTile devices timestamp their readings in uptime rather than UTC time,
which means a conversion process is needed to convert to absolute time from
the device's local reference time.

The UTCAssigner class provides a reference implementation of this conversion
process.  The conversion algorithm looks for known points in time where both
the device uptime and absolute UTC time are known.  These points are known as
anchors.  For any other reading, the difference in uptime between it and the
closest anchor is added to the absolute UTC time of the anchor to calculate
the absolute UTC time of the reading.

In the easy case, when there are no device resets between the reading you want
to convert and an anchor point and uptimes are known for each point, this
reconstruction process is exact.

However, in the general case, an exact assignment is not possible and
UTCAssigner uses various other methods to infer an approximate UTC timestamp,
returning confidence metrics along with the assigned value.
"""

import datetime
import logging
from typedargs.exceptions import ArgumentError
from sortedcontainers import SortedKeyList
from .signed_list_format import SignedListReport
from .report import IOTileReading


class _TimeAnchor:
    """Internal class for storing a utc reference point."""

    __slots__ = ('uptime', 'utc', 'reading_id', 'is_break', 'exact')

    def __init__(self, reading_id, uptime=None, utc=None, is_break=False, exact=False):
        self.uptime = uptime
        self.utc = utc
        self.reading_id = reading_id
        self.is_break = is_break
        self.exact = exact

    def copy(self):
        """Return a copy of this _TimeAnchor."""
        return _TimeAnchor(self.reading_id, self.uptime, self.utc, self.is_break, self.exact)


class UTCAssignment:
    """Data class recording the assignment of a UTC timestamp.

    Args:
        reading_id (int): The id of the reading that we are assigning
        utc (datetime): UTC datetime that we assigned to the reading
        found (bool): Whether we found the exact reading_id in our
            set of anchor points or inferred it as between two other
            reading ids
        exact (bool): Whether we believe the UTC assignment is exact
            or approximate.
        crossed_break (bool): Whether the assignment required crossing
            an uptime break such as a device reset.  This can introduce
            a lot of ambiguity into the result since the time-break could
            be arbitrarily long.
    """

    _Y2KReference = datetime.datetime(2000, 1, 1)
    _EpochReference = datetime.datetime(1970, 1, 1)

    __slots__ = ('reading_id', 'utc', 'found_id', 'exact', 'crossed_break', 'rtc_value')

    def __init__(self, reading_id, utc, found, exact, crossed_break):
        self.reading_id = reading_id
        self.utc = utc
        self.found_id = found
        self.exact = exact
        self.crossed_break = crossed_break

        rtc_value = int((utc - self._Y2KReference).total_seconds())
        self.rtc_value = rtc_value | (1 << 31)

    @property
    def epoch_value(self):
        """The number of seconds since the Unix Epoch."""

        return int((self.utc - self._EpochReference).total_seconds())

    def __str__(self):
        return "%s (reading_id=%08X, exact=%s, crossed_break=%s)" % \
               (self.utc, self.reading_id, self.exact, self.crossed_break)


class UTCAssigner:
    """Helper class to assign UTC timestamps to device readings.

    This class contains logic to infer UTC timestamps for readings that are
    stamped in uptime only by looking for nearby readings for which the UTC
    timestamp is known.  The relative offset between these anchor points and
    the reading in question is then used to infer the UTC timestamp from the
    anchor point either exactly or approximately.

    The underlying truth that this class relies on is the fact that all
    readings produced by an IOTile device have a mononotically increasing
    reading_id that can be used to absolutely order them.  In contrast to the
    readings timestamp, which may reset to 0 if the device reboots and does
    not have a realtime clock, the reading_id is guaranteed to never decrease.
    """

    _Y2KReference = datetime.datetime(2000, 1, 1)
    _EpochReference = datetime.datetime(1970, 1, 1)

    def __init__(self):
        self._anchor_points = SortedKeyList(key=lambda x: x.reading_id)
        self._prepared = False
        self._anchor_streams = {}
        self._break_streams = set()
        self._logger = logging.getLogger(__name__)

        self._known_converters = {
            'rtc': UTCAssigner._convert_rtc_anchor,
            'epoch': UTCAssigner._convert_epoch_anchor
        }

    def _load_known_breaks(self):
        self._break_streams.add(0x5C00)

    def anchor_stream(self, stream_id, converter="rtc"):
        """Mark a stream as containing anchor points."""

        if isinstance(converter, str):
            converter = self._known_converters.get(converter)

            if converter is None:
                raise ArgumentError("Unknown anchor converter string: %s" % converter,
                                    known_converters=list(self._known_converters))

        self._anchor_streams[stream_id] = converter

    def id_range(self):
        """Get the range of archor reading_ids.

        Returns:
            (int, int): The lowest and highest reading ids.

            If no reading ids have been loaded, (0, 0) is returned.
        """

        if len(self._anchor_points) == 0:
            return (0, 0)

        return (self._anchor_points[0].reading_id, self._anchor_points[-1].reading_id)

    @classmethod
    def convert_rtc(cls, timestamp):
        """Convert a number of seconds since 1/1/2000 to UTC time."""

        if timestamp & (1  << 31):
            timestamp &= ~(1 << 31)

        delta = datetime.timedelta(seconds=timestamp)
        return cls._Y2KReference + delta

    @classmethod
    def _convert_rtc_anchor(cls, reading):
        """Convert a reading containing an RTC timestamp to datetime."""

        return cls.convert_rtc(reading.value)

    @classmethod
    def _convert_epoch_anchor(cls, reading):
        """Convert a reading containing an epoch timestamp to datetime."""

        delta = datetime.timedelta(seconds=reading.value)
        return cls._EpochReference + delta

    def add_point(self, reading_id, uptime=None, utc=None, is_break=False):
        """Add a time point that could be used as a UTC reference."""

        if reading_id == 0:
            return

        if uptime is None and utc is None:
            return

        if uptime is not None and uptime & (1 << 31):
            if utc is not None:
                return

            uptime &= ~(1 << 31)

            utc = self.convert_rtc(uptime)
            uptime = None

        anchor = _TimeAnchor(reading_id, uptime, utc, is_break, exact=utc is not None)

        if anchor in self._anchor_points:
            return

        self._anchor_points.add(anchor)
        self._prepared = False

    def add_reading(self, reading):
        """Add an IOTileReading."""

        is_break = False
        utc = None

        if reading.stream in self._break_streams:
            is_break = True

        if reading.stream in self._anchor_streams:
            utc = self._anchor_streams[reading.stream](reading)

        self.add_point(reading.reading_id, reading.raw_time, utc, is_break=is_break)

    def add_report(self, report, ignore_errors=False):
        """Add all anchors from a report."""

        if not isinstance(report, SignedListReport):
            if ignore_errors:
                return

            raise ArgumentError("You can only add SignedListReports to a UTCAssigner", report=report)

        for reading in report.visible_readings:
            self.add_reading(reading)

        self.add_point(report.report_id, report.sent_timestamp, report.received_time)

    def assign_utc(self, reading_id, uptime=None, prefer="before"):
        """Assign a utc datetime to a reading id.

        This method will return an object with assignment information or None
        if a utc value cannot be assigned.  The assignment object returned
        contains a utc property that has the asssigned UTC as well as other
        properties describing how reliable the assignment is.

        Args:
            reading_id (int): The monotonic reading id that we wish to assign
                a utc timestamp to.
            uptime (int): Optional uptime that should be associated with the
                reading id.  If this is not specified and the reading_id is
                found in the anchor points passed to this class then the
                uptime from the corresponding anchor point will be used.
            prefer (str): There are two possible directions that can be used
                to assign a UTC timestamp (the nearest anchor before or after the
                reading).  If both directions are of similar quality, the choice
                is arbitrary.  Passing prefer="before" will use the anchor point
                before the reading.  Passing prefer="after" will use the anchor
                point after the reading.  Default: before.

        Returns:
            UTCAssignment: The assigned UTC time or None if assignment is impossible.
        """

        if prefer not in ("before", "after"):
            raise ArgumentError("Invalid prefer parameter: {}, must be 'before' or 'after'".format(prefer))

        if len(self._anchor_points) == 0:
            return None

        if reading_id > self._anchor_points[-1].reading_id:
            return None

        i = self._anchor_points.bisect_key_left(reading_id)
        found_id = False
        crossed_break = False
        exact = True

        last = self._anchor_points[i].copy()
        if uptime is not None:
            last.uptime = uptime

        if last.reading_id == reading_id:
            found_id = True

            if last.utc is not None:
                return UTCAssignment(reading_id, last.utc, found_id, exact, crossed_break)

        left_assign = self._fix_left(reading_id, last, i, found_id)
        if left_assign is not None and left_assign.exact:
            return left_assign

        right_assign = self._fix_right(reading_id, last, i, found_id)
        if right_assign is not None and right_assign.exact:
            return right_assign

        return self._pick_best_fix(left_assign, right_assign, prefer)

    def ensure_prepared(self):
        """Calculate and cache UTC values for all exactly known anchor points."""

        if self._prepared:
            return

        exact_count = 0
        fixed_count = 0
        inexact_count = 0

        self._logger.debug("Preparing UTCAssigner (%d total anchors)", len(self._anchor_points))

        for curr in self._anchor_points:
            if not curr.exact:
                assignment = self.assign_utc(curr.reading_id, curr.uptime)
                if assignment is not None and assignment.exact:
                    curr.utc = assignment.utc
                    curr.exact = True
                    fixed_count += 1
                else:
                    inexact_count += 1
            else:
                exact_count += 1

        self._logger.debug("Prepared UTCAssigner with %d reference points, "
                           "%d exact anchors and %d inexact anchors",
                           exact_count, fixed_count, inexact_count)

        self._prepared = True

    def fix_report(self, report, errors="drop", prefer="before"):
        """Perform utc assignment on all readings in a report.

        The returned report will have all reading timestamps in UTC. This only
        works on SignedListReport objects.  Note that the report should
        typically have previously been added to the UTC assigner using
        add_report or no reference points from the report will be used.

        Args:
            report (SignedListReport): The report that we should fix.
            errors (str): The behavior that we should have when we can't
                fix a given reading.  The only currently support behavior is
                drop, which means that the reading will be dropped and not
                included in the new report.
            prefer (str): Whether to prefer fixing readings by looking for
                reference points after the reading or before, all other things
                being equal.  See the description of ``assign_utc``.

        Returns:
            SignedListReport: The report with UTC timestamps.
        """

        if not isinstance(report, SignedListReport):
            raise ArgumentError("Report must be a SignedListReport", report=report)

        if errors not in ('drop',):
            raise ArgumentError("Unknown errors handler: {}, supported=['drop']".format(errors))

        self.ensure_prepared()

        fixed_readings = []
        dropped_readings = 0

        for reading in report.visible_readings:
            assignment = self.assign_utc(reading.reading_id, reading.raw_time, prefer=prefer)

            if assignment is None:
                dropped_readings += 1
                continue

            fixed_reading = IOTileReading(assignment.rtc_value, reading.stream, reading.value,
                                          reading_time=assignment.utc, reading_id=reading.reading_id)
            fixed_readings.append(fixed_reading)

        fixed_report = SignedListReport.FromReadings(report.origin, fixed_readings, report_id=report.report_id,
                                                     selector=report.streamer_selector, streamer=report.origin_streamer,
                                                     sent_timestamp=report.sent_timestamp)
        fixed_report.received_time = report.received_time

        if dropped_readings > 0:
            self._logger.warning("Dropped %d readings of %d when fixing UTC timestamps in report 0x%08X for device 0x%08X",
                                 dropped_readings, len(report.visible_readings), report.report_id, report.origin)

        return fixed_report

    def _pick_best_fix(self, before, after, prefer):
        if before is None and after is None:
            return None

        if after is None:
            return before

        if before is None:
            return after

        if after.crossed_break and not before.crossed_break:
            return before

        if before.crossed_break and not after.crossed_break:
            return after

        if before.exact and not after.exact:
            return before

        if after.exact and not before.exact:
            return after

        if prefer == 'before':
            return before

        return after

    def _fix_right(self, reading_id, last, start, found_id):
        """Fix a reading by looking for the nearest anchor point after it."""

        accum_delta = 0
        exact = True
        crossed_break = False

        if start == len(self._anchor_points) - 1:
            return None

        for curr in self._anchor_points.islice(start + 1):
            if curr.uptime is None or last.uptime is None:
                exact = False
            elif curr.is_break or curr.uptime < last.uptime:
                exact = False
                crossed_break = True
            else:
                accum_delta += curr.uptime - last.uptime

            if curr.utc is not None:
                time_delta = datetime.timedelta(seconds=accum_delta)
                return UTCAssignment(reading_id, curr.utc - time_delta, found_id, exact, crossed_break)

            last = curr

        return None

    def _fix_left(self, reading_id, last, start, found_id):
        """Fix a reading by looking for the nearest anchor point before it."""

        accum_delta = 0
        exact = True
        crossed_break = False

        if start == 0:
            return None

        for curr in self._anchor_points.islice(None, start, reverse=True):
            if curr.uptime is None or last.uptime is None:
                exact = False
            elif curr.is_break or last.uptime < curr.uptime:
                exact = False
                crossed_break = True
            else:
                accum_delta += last.uptime - curr.uptime

            if curr.utc is not None:
                time_delta = datetime.timedelta(seconds=accum_delta)
                return UTCAssignment(reading_id, curr.utc + time_delta, found_id, exact, crossed_break)

            last = curr

        return None



"""A wrapper class that turns a legacy DeviceAdapter into an AbstractDeviceAdapter.

The legacy adapter routines are run in a thread executor to provide a
non-blocking coroutine based api.  All callbacks from the legacy adapter are
translated to the corresponding event type appropriate and notified as events.

The periodic callback from the legacy adapter is called in a loop once per second
with the the periodic_callback itself running in an executor thread in case it
contains blocking operations.
"""

import asyncio
import logging
import functools
import concurrent.futures
from iotile.core.exceptions import ArgumentError
from iotile.core.utilities import SharedLoop
from iotile.core.hw.reports import BroadcastReport
from .standard import StandardDeviceAdapter
from .legacy import DeviceAdapter
from ...exceptions import DeviceAdapterError
from ...virtual import unpack_rpc_response

_MISSING = object()


class AsynchronousModernWrapper(StandardDeviceAdapter):
    """Provide a modern coroutine API on top of a legacy DeviceAdapter."""

    def __init__(self, adapter, loop=SharedLoop):
        super(AsynchronousModernWrapper, self).__init__(loop=loop)

        if not isinstance(adapter, DeviceAdapter):
            raise ArgumentError("AsynchronousModernWrapper created from object "
                                "not a subclass of DeviceAdapter", adapter=adapter)

        self._adapter = adapter
        self._loop = loop
        self._logger = logging.getLogger(__name__)
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)
        self._task = loop.add_task(None, name="LegacyAdapter (%s)" % adapter.__class__.__name__,
                                   finalizer=self.stop, stop_timeout=5.0)

    def set_config(self, name, value):
        """Set a config value for this adapter by name

        Args:
            name (string): The name of the config variable
            value (object): The value of the config variable
        """

        self._adapter.set_config(name, value)

    def get_config(self, name, default=_MISSING):
        """Get a config value from this adapter by name

        Args:
            name (string): The name of the config variable
            default (object): The default value to return if config is not found

        Returns:
            object: the value associated with the name

        Raises:
            ArgumentError: if the name is not found and no default is supplied
        """

        value = self._adapter.get_config(name, default)
        if value is _MISSING:
            raise ArgumentError("Config value did not exist", name=name)

        return value

    async def start(self):
        """Start the device adapter.

        See :meth:`AbstractDeviceAdapter.start`.
        """

        self._loop.add_task(self._periodic_loop, name="periodic task for %s" % self._adapter.__class__.__name__,
                            parent=self._task)

        self._adapter.add_callback('on_scan', functools.partial(_on_scan, self._loop, self))
        self._adapter.add_callback('on_report', functools.partial(_on_report, self._loop, self))
        self._adapter.add_callback('on_trace', functools.partial(_on_trace, self._loop, self))
        self._adapter.add_callback('on_disconnect', functools.partial(_on_disconnect, self._loop, self))

    async def stop(self, _task=None):
        """Stop the device adapter.

        See :meth:`AbstractDeviceAdapter.stop`.
        """

        self._logger.info("Stopping adapter wrapper")

        if self._task.stopped:
            return

        for task in self._task.subtasks:
            await task.stop()

        self._logger.debug("Stopping underlying adapter %s", self._adapter.__class__.__name__)
        await self._execute(self._adapter.stop_sync)

    def _execute(self, func, *args):
        return self._loop.get_loop().run_in_executor(self._executor, func, *args)

    async def _periodic_loop(self):
        while True:
            try:
                await self._execute(self._adapter.periodic_callback)
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                raise
            except:  #pylint:disable=bare-except;This is a background worker task that logs the error
                self._logger.exception("Error in periodic callback from device adapter")

    def can_connect(self):
        """Return whether this device adapter can accept another connection.

        We just forward the request to the underlying adapter and ask.

        See :meth:`AbstractDeviceAdapter.can_connect`.
        """

        return self._adapter.can_connect()

    async def connect(self, conn_id, connection_string):
        """Connect to a device.

        See :meth:`AbstractDeviceAdapter.connect`.
        """

        self._logger.info("Inside connect, conn_id=%d, conn_string=%s", conn_id, connection_string)

        try:
            self._setup_connection(conn_id, connection_string)

            resp = await self._execute(self._adapter.connect_sync, conn_id, connection_string)
            _raise_error(conn_id, 'connect', resp)
        except:
            self._teardown_connection(conn_id, force=True)
            raise

    async def disconnect(self, conn_id):
        """Disconnect from a connected device.

        See :meth:`AbstractDeviceAdapter.disconnect`.
        """

        resp = await self._execute(self._adapter.disconnect_sync, conn_id)
        _raise_error(conn_id, 'disconnect', resp)
        self._teardown_connection(conn_id, force=True)

    async def open_interface(self, conn_id, interface):
        """Open an interface on an IOTile device.

        See :meth:`AbstractDeviceAdapter.open_interface`.
        """

        resp = await self._execute(self._adapter.open_interface_sync, conn_id, interface)
        _raise_error(conn_id, 'open_interface', resp)

    async def close_interface(self, conn_id, interface):
        """Close an interface on this IOTile device.

        See :meth:`AbstractDeviceAdapter.close_interface`.
        """

        resp = await self._execute(self._adapter.close_interface_sync, conn_id, interface)
        _raise_error(conn_id, 'close_interface', resp)

    async def probe(self):
        """Probe for devices connected to this adapter.

        See :meth:`AbstractDeviceAdapter.probe`.
        """

        resp = await self._execute(self._adapter.probe_sync)
        _raise_error(None, 'probe', resp)

    async def send_rpc(self, conn_id, address, rpc_id, payload, timeout):
        """Send an RPC to a device.

        See :meth:`AbstractDeviceAdapter.send_rpc`.
        """

        resp = await self._execute(self._adapter.send_rpc_sync, conn_id, address, rpc_id, payload, timeout)
        _raise_error(conn_id, 'send_rpc', resp)

        status = resp.get('status')
        payload = resp.get('payload')

        # This will raise an exception if needed based on status
        return unpack_rpc_response(status, payload, rpc_id, address)

    async def debug(self, conn_id, name, cmd_args):
        """Send a debug command to a device.

        See :meth:`AbstractDeviceAdapter.debug`.
        """

        progress_callback = functools.partial(_on_progress, self, 'debug', conn_id)

        resp = await self._execute(self._adapter.debug_sync, conn_id, name, cmd_args, progress_callback)
        _raise_error(conn_id, 'send_rpc', resp)

        return resp.get('return_value')

    async def send_script(self, conn_id, data):
        """Send a a script to a device.

        See :meth:`AbstractDeviceAdapter.send_script`.
        """

        progress_callback = functools.partial(_on_progress, self, 'script', conn_id)

        resp = await self._execute(self._adapter.send_script_sync, conn_id, data, progress_callback)
        _raise_error(conn_id, 'send_rpc', resp)


def _raise_error(conn_id, operation, resp):
    if resp['success'] is False:
        raise DeviceAdapterError(conn_id, operation, resp.get('failure_reason', "unknown failure reason"))


def _on_scan(_loop, adapter, _adapter_id, info, expiration_time):
    """Callback when a new device is seen."""

    info['validity_period'] = expiration_time
    adapter.notify_event_nowait(info.get('connection_string'), 'device_seen', info)


def _on_report(_loop, adapter, conn_id, report):
    """Callback when a report is received."""

    conn_string = None
    if conn_id is not None:
        conn_string = adapter._get_property(conn_id, 'connection_string')

    if isinstance(report, BroadcastReport):
        adapter.notify_event_nowait(conn_string, 'broadcast', report)
    elif conn_string is not None:
        adapter.notify_event_nowait(conn_string, 'report', report)
    else:
        adapter._logger.debug("Dropping report with unknown conn_id=%s", conn_id)


def _on_trace(_loop, adapter, conn_id, trace):
    """Callback when tracing data is received."""

    conn_string = adapter._get_property(conn_id, 'connection_string')
    if conn_string is None:
        adapter._logger.debug("Dropping trace data with unknown conn_id=%s", conn_id)
        return

    adapter.notify_event_nowait(conn_string, 'trace', trace)


def _on_disconnect(_loop, adapter, _adapter_id, conn_id):
    """Callback when a device disconnects unexpectedly."""

    conn_string = adapter._get_property(conn_id, 'connection_string')
    if conn_string is None:
        adapter._logger.debug("Dropping disconnect notification with unknown conn_id=%s", conn_id)
        return

    adapter._teardown_connection(conn_id, force=True)
    event = dict(reason='no reason passed from legacy adapter', expected=False)
    adapter.notify_event_nowait(conn_string, 'disconnection', event)


def _on_progress(adapter, operation, conn_id, done, total):
    """Callback when progress is reported."""

    conn_string = adapter._get_property(conn_id, 'connection_string')
    if conn_string is None:
        return

    adapter.notify_progress(conn_string, operation, done, total)



from typedargs.exceptions import ArgumentError
from .proxy import TileBusProxyObject


class TileBusProxyPlugin:
    """
    Proxy plugin objects are mixin style classes that add functionality modules to proxy objects
    """

    def __init__(self, parent):
        if not isinstance(parent, TileBusProxyObject):
            raise ArgumentError("Attempting to initialize a TileBusProxyPlugin with an invalid parent object",
                                parent=parent)

        self._proxy = parent

    def rpc(self, feature, cmd, *args, **kw):
        return self._proxy.rpc(feature, cmd, *args, **kw)

    def rpc_v2(self, cmd, arg_format, result_format, *args, **kwargs):
        return self._proxy.rpc_v2(cmd, arg_format, result_format, *args, **kwargs)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

# exceptions.py

import iotile.core.exceptions

class DeviceAdapterError(iotile.core.exceptions.HardwareError):
    """A protocol or communication error occurred during an operation.

    This exception indicates that a DeviceAdapter was not able to perform
    the desired operation because of a communication-type error.

    Args:
        conn_id (int): The connection id that the error occurred on.  If the
            error was not associated with a connection then this will be None.
        operation (str): A descriptive name of the operation that was being
            performed.
        reason (str): A helpful string about what the issue was.
    """

    def __init__(self, conn_id, operation, reason):
        super(DeviceAdapterError, self).__init__("Operation {} on conn {} failed because {}".
                                                 format(operation, conn_id, reason),
                                                 reason=reason, operation=operation,
                                                 conn_id=conn_id)

        self.reason = reason


class DeviceServerError(iotile.core.exceptions.HardwareError):
    """The device server rejected your request because it was invalid.

    This exception indicates that the request was never forwarded to the
    underlying device adapter but rather reject by the DeviceServer
    implementation itself because it was disallowed or invalid.

    For example, if you are trying to access a device for which you
    do not have an active connection.  Or disconnecting a device that
    you did not connect.

    The client_id, reason and connection_string are available as
    properties on the exception instance to assist with converting
    the error into a message to send back to the client.

    Args:
        client_id (str): The client id that requested the invalid
            operation.
        conn_str (str): The connection string for the device in question.
        operation (str): A descriptive name of the operation that was
            requested.
        reason (str): A helpful string about what the issue was.
    """

    def __init__(self, client_id, conn_str, operation, reason):
        super(DeviceServerError, self).__init__("Operation {} for client {} on device {} failed because {}".
                                                format(operation, client_id, conn_str, reason),
                                                reason=reason, operation=operation,
                                                conn_string=conn_str, client_id=client_id)

        self.reason = reason
        self.client_id = client_id
        self.connection_string = conn_str


class DevicePushError(iotile.core.exceptions.HardwareError):
    """Base class for errors related to virtual devices pushing to clients.

    These errors can be thrown when a virtual device attempts to stream
    reports and the report streaming process fails or when it attempts to
    trace data.

    The details of the failure reason are contained in the message.
    """


class InterfaceClosedError(DevicePushError):
    """An asynchronous device push failed because the required interface was closed.

    This is a typical error that is raised when, for example, a device
    attempts to stream a report without the streaming interface being opened.
    """

    def __init__(self, interface):
        super(InterfaceClosedError, self).__init__("Cannot push data to client because the {} interface is not open".
                                                   format(interface), interface=interface)


class RPCError(iotile.core.exceptions.HardwareError):
    """Base class for all RPC related errors.

    Instances of this class should not be raised directly but it
    just provides a standard base for checking against the specific
    kinds of errors that can happen while running an RPC.
    """


class RPCNotFoundError(RPCError):
    """The RPC id was not found on the tile.

    This exception always indicates that the tile itself was responding
    correctly and was able to handle RPCs.  It just did not find an RPC
    matching the RPC id that was specified.
    """


class RPCInvalidArgumentsError(RPCError):
    """An RPC with a fixed parameter format was given invalid arguments."""


class RPCInvalidReturnValueError(RPCError):
    """The return value of an RPC does not conform to its known format.

    When the RPC is implemented in python in a virtual tile, this may also be
    thrown if invalid python objects are returned that cannot be packed for
    transport.
    """

    def __init__(self, address, rpc_id, format_code, response, **kwargs):
        if response is None:
            resp_len = 0
        else:
            resp_len = len(response)

        super(RPCInvalidReturnValueError, self).__init__("Invalid return value from Tile %s, RPC 0x%04x, "
                                                         "expected format %s, got %d bytes"
                                                         % (address, rpc_id, format_code, resp_len), **kwargs)

        self.address = address
        self.rpc_id = rpc_id
        self.format_code = format_code
        self.response = response


class RPCInvalidIDError(RPCError):
    """The RPC ID given is an invalid type or size to specify an RPC."""


class TileNotFoundError(RPCError):
    """An RPC was sent to a tile that does not exist.

    This is distinct from :class:`RPCNotFoundError`, which means that the
    destination address was correct and that tile responded positively that
    the RPC could not be found.  This error means that no response was
    received from the tile at all.
    """


class RPCErrorCode(RPCError):
    """The RPC chose to return an implementation defined error code.

    Use of this exception is discouraged since RPCs implementations should
    return their error codes inside the response payload along with any data
    that they may want to return.  However there are certain cases where RPC
    implementation need a way to indicate an error while returning no response
    and raising this RPCErrorCode exception is the way to do it.

    Args:
        status_code (int): The error code that should be raised to the caller.
    """

    def __init__(self, status_code):
        super(RPCErrorCode, self).__init__("RPC returned application defined status code %d" % status_code,
                                           code=status_code)


class AsynchronousRPCResponse(RPCError):
    """Exception thrown from an RPC implementation when it will return asynchronously.

    This exception is never returned from a call to self.rpc or self.rpc_v2
    inside of a TileBus proxy, it is just used internally to know when to
    block that call and wait for the actual RPC response to come through a
    callback.
    """

    def __init__(self):
        super(AsynchronousRPCResponse, self).__init__("RPC handler elected to return asynchronously")


class BusyRPCResponse(RPCError):
    """The Tile was busy and could not handle an RPC right now.

    This exception is rarely exposed directly to a user.  Internally, it can be
    the case that an RPC cannot be executed at a specific moment because a
    long-running asynchronous RPC is in progress.  In that case, the
    :meth:`TileBusProxyObject.rpc` method automatically retries the RPC later.
    If multiple attempts to retry the RPC fail, this exception will ultimately
    be raised to the user.
    """

    def __init__(self, msg="Tile tile is busy"):
        super(BusyRPCResponse, self).__init__(msg)


VALID_RPC_EXCEPTIONS = (BusyRPCResponse, TileNotFoundError, RPCErrorCode, RPCInvalidIDError, RPCNotFoundError)


class UnknownModuleTypeError(iotile.core.exceptions.TypeSystemError):
    """A suitable proxy could not be found for a given tile.

    This means that the name returned by the tile's get_info RPC was not found
    in our local database of known tile names so a suitable proxy class could
    not be instantiated.  Typically this means that you need to find and
    install the support package for the tile that you wish to use.
    """



"""A general purpose virtual device base class.

This base class supports adding tiles to your virtual device containing RPCs.
Each tile is assigned an address and can have many RPCs tagged with
``@tile_rpc`` decorators.

Unlike :class:`SimpleVirtualDevice`, there is no way to define an RPC without
putting it inside of a tile first.  This makes this class slightly more
complicated to use when you only want to implement a very small number of RPCs
and are not interested in reusing your RPC implementations.  For those cases,
:class:`SimpleVirtualDevice` is a better base class to inherit from.
"""

import inspect
from iotile.core.exceptions import ArgumentError
from iotile.core.utilities import SharedLoop
from ..exceptions import RPCInvalidIDError, RPCNotFoundError, TileNotFoundError
from .virtualdevice_base import BaseVirtualDevice


class StandardVirtualDevice(BaseVirtualDevice):
    """A basic virtual device base class with support for tiles and scripts.

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        loop (BackgroundEventLoop): The loop we should use for running background
            tasks. Defaults to the global SharedLoop.
    """

    __NO_EXTENSION__ = True

    def __init__(self, iotile_id, *, loop=SharedLoop):
        BaseVirtualDevice.__init__(self, iotile_id, loop=loop)

        self._tiles = {}
        self.script = bytearray()

    def push_script_chunk(self, chunk):
        """Called when someone pushes a new bit of a TRUB script to this device

        Args:
            chunk (str): a buffer with the next bit of script to append
        """

        self.script += bytearray(chunk)

    def add_tile(self, address, tile):
        """Add a tile to handle all RPCs at a given address.
        Args:
            address (int): The address of the tile
            tile (RPCDispatcher): A tile object that inherits from RPCDispatcher
        """

        if address in self._tiles:
            raise ArgumentError("Tried to add two tiles at the same address", address=address)

        self._tiles[address] = tile

    async def async_rpc(self, address, rpc_id, payload=b""):
        """Call an RPC by its address and ID.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes
        Returns:
            bytes: The response payload from the RPC
        """

        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if address not in self._tiles:
            raise TileNotFoundError("Unknown tile address, no registered handler", address=address)

        tile = self._tiles.get(address)
        if tile is not None and tile.has_rpc(rpc_id):
            resp = tile.call_rpc(rpc_id, payload)
            if inspect.isawaitable(resp):
                resp = await resp

            return resp

        raise RPCNotFoundError("Could not find RPC 0x%X at address %d" % (rpc_id, address))



"""Base class for software based iotile devices called virtual devices."""

import abc
import inspect
import logging
from iotile.core.exceptions import InternalError, ArgumentError
from iotile.core.utilities import SharedLoop
from ..exceptions import RPCNotFoundError, DevicePushError
from ..reports.individual_format import IndividualReadingReport
from ..reports.report import IOTileReading
from .common_types import RPCDeclaration, pack_rpc_payload, unpack_rpc_payload


class AbstractAsyncDeviceChannel(abc.ABC):
    """Abstract class for virtual devices to push events to clients.

    When a virtual device is started, it is passed a subclass of this base
    class in its :meth:`BaseVirtualDevice.start` method and it can use that
    class to push reports via the streaming interface, tracing data via the
    tracing interface, broadcast data or to disconnect the client forcibly.

    This base class just defines the methods that must be provided and their
    signatures.  Each method must return an awaitable object.
    """

    @abc.abstractmethod
    def stream(self, report):
        """Stream a report to the client.

        This method must yield until the report has been successfully sent or
        queued for sending in order to propogate backpressure to callers that
        are streaming reports faster than the DeviceServer can actually send
        them.

        If the report is not able to be sent, an instance of :class:`StreamingError`
        must be raised.

        Args:
            report (IOTileReport): The report that should be streamed out of the
                device.  If the report is a subclass of :class:`BroadcastReport`,
                it will be broadcast instead of streamed.

        Returns:
            awaitable: An awaitable that will finish when the report has been sent.

            Note that there is no gurantee that the report has succesfully
            reached the client when this method returns since many things
            could have gone wrong in the air or on the client's software.
            This response is purely useful for flow control.

        Raises:
            DevicePushError: The report could not be streamed/queued for streaming.
        """

    @abc.abstractmethod
    def trace(self, data):
        """Send binary tracing data to the client.

        This method must yield until the tracing data is successfully sent or
        queued for sending in order to propogate backpressure to callers that
        are tracing data faster than the DeviceServer can actually send it.

        If the data cannot be sent or queued, an instance of :class:`TracingError`
        must be raised.

        Args:
            data (bytes): The raw data that should be sent out of the tracing interface

        Returns:
            awaitable: An awaitable that will finish when the tracing data has been sent.

            Note that there is no gurantee that the data has succesfully
            reached the client when this method returns since many things
            could have gone wrong in the air or on the client's software. This
            response is purely useful for flow control.

        Raises:
            DevicePushError: The tracing data could not be send/queued for sending.
        """

    @abc.abstractmethod
    def disconnect(self):
        """Forcibly disconnect the connected client.

        In order to prevent race conditions, this method should not return
        until the client has been disconnected.

        Returns:
            awaitable: An awaitable that resolves when the client has been disconnected.

        Raises:
            DevicePushError: If there is no client connected or they could not be disconnected.
        """


# pylint: disable=R0902,R0904; backwards compatibility methods and properties already referenced in other modules
class BaseVirtualDevice:
    """A Virtual IOTile device that can be interacted with as if it were a real one.

    This is the base class of all other Virtual IOTile devices.

    This base class does not make any assumptions about how your particular
    subclass is organized and as such is unlikely to be the best parent class
    for any particular use case.  Depending on what you want to do, you may
    want one of the following subclasses:

    - :class:`StandardVirtualDevice`: A general purpose base class for complex
      virtual devices supporting organizing the device into tiles.

    - :class:`SimpleVirtualDevice`: A simpler base class for creating very
      simple virtual devices that just implement a few RPCs without needing to
      put those rpcs into tiles.

    .. important::

        All subclasses must override the function :meth:`call_rpc` to actually
        find and call their RPCs, otherwise the default impementation will
        raise ``RPCNotFoundError`` for all RPCs.  The above two subclasses
        provide appropriate implementations for ``call_rpc``.

    In order to be able to dynamically created by :class:`VirtualDeviceAdapter`,
    all BaseVirtualDevice subclasses must have the following __init__ signature::

        __init__(self, config, *, loop=SharedLoop)

    The ``config`` parameter is a dictionary of arguments that can be used in
    any way the virtual device class likes to configure itself.  If the device
    needs explicit access to a BackgroundEventLoop, it can declare a keyword
    parameter named ``loop``, which will be filled in with the appropriate
    ``BackgroundEventLoop`` that the ``VirtualDeviceAdapter`` is using
    internally.

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        loop (BackgroundEventLoop): The event loop to use to run coroutines.
    """

    def __init__(self, iotile_id, loop=SharedLoop):
        self.iotile_id = iotile_id

        self._interface_status = dict(connected=False, streaming=False, rpc=False,
                                      tracing=False, script=False, debug=False)

        # For this device to push streams or tracing data through a VirtualInterface, it
        # needs access to that interface's push channel
        self._push_channel = None
        self._started = False
        self._logger = logging.getLogger(__name__)
        self._loop = loop

    @property
    def connected(self):
        """Whether someone is connected to the virtual device."""

        return self._interface_status['connected']

    def interface_open(self, name):
        """Check whether the given interface is open.

        Valid interface names are:
         - connected
         - stream
         - trace
         - script
         - debug

        Interfaces are opened by calling open_interface and closed by calling
        close_interface, except for connections which are opened by setting
        the connect propery and closed by clearing the connect property.

        Args:
            name (str): The name of the interface.

        Returns:
            bool: Whether the interface is open.
        """

        if name not in self._interface_status:
            raise ArgumentError("Unkown interface name: %s" % name)

        return self._interface_status[name]

    def start(self, channel):
        """Start running this virtual device.

        Args:
            channel (AbstractAsyncDeviceChannel): A channel that can be used by the device
                to push events asynchronously to an attached client.
        """

        if self._started:
            raise InternalError("The method start() was called twice on VirtualIOTileDevice.")

        self._push_channel = channel
        self._started = True

    def stop(self):
        """Stop running this virtual device."""

        if not self._started:
            raise InternalError("Stop called without start() first being called")

        self._started = False

    async def connect(self):
        """Connect to this virtual device."""

        if self.connected:
            raise InternalError("connect() was called but someone was already connected")

        await self.open_interface('connected')

    async def disconnect(self):
        """Disconnect from this virtual device.

        Any other interfaces that are opened on the device will be closed
        automatically before the disconnect routine completes.  After this
        method finishes successfully, you will be able to connect to the
        device again.
        """

        if not self.connected:
            raise InternalError("disconnect() was called without a connection")

        for key in ('streaming', 'tracing', 'debug', 'script', 'rpc'):
            if self.interface_open(key):
                await self.close_interface(key)

        await self.close_interface('connected')

    async def stream(self, report):
        """Stream a report asynchronously.

        If no one is listening for the report, the report may be dropped,
        otherwise it will be queued for sending

        Args:
            report (IOTileReport): The report that should be streamed
            callback (callable): Optional callback to get notified when
                this report is actually sent.

        Returns:
            awaitable: Resolves when the streaming finishes
        """

        if self._push_channel is None:
            raise DevicePushError("No push channel configured for device")

        await self._push_channel.stream(report)

    async def stream_realtime(self, stream, value):
        """Stream a realtime value as an IndividualReadingReport.

        If the streaming interface of the VirtualInterface this
        VirtualDevice is attached to is not opened, the realtime
        reading may be dropped.

        Args:
            stream (int): The stream id to send
            value (int): The stream value to send
        """

        if not self.interface_open('streaming'):
            return

        reading = IOTileReading(0, stream, value)

        report = IndividualReadingReport.FromReadings(self.iotile_id, [reading])
        self.stream(report)

    async def trace(self, data):
        """Trace data asynchronously.

        If no one is listening for traced data, it will be dropped
        otherwise it will be queued for sending.

        Args:
            data (bytes): Unstructured data to trace to any
                connected client.
        """

        if self._push_channel is None:
            raise DevicePushError("No push channel configured for device")

        await self._push_channel.trace(data)

    async def open_interface(self, name):
        """Open an interface on the device by name.

        Subclasses that want to be notified when an interface is opened can
        declare a method named open_{name}_interface that will be called when
        the given interface is opened.  If such a function is delcared and
        raises an exception, the interface will not be opened.

        Args:
            name (str): The name of the interface to open.
        """

        if name not in self._interface_status:
            raise ArgumentError("Unknown interface name in open_interface: {}".format(name))

        self._interface_status[name] = True

        try:
            iface_open_name = "_open_{}_interface".format(name)

            if hasattr(self, iface_open_name):
                open_func = getattr(self, iface_open_name)
                res = open_func()
                if inspect.isawaitable(res):
                    res = await res

                return res

            return None
        except:
            self._interface_status[name] = False
            raise

    async def close_interface(self, name):
        """Close an interface on the device by name.

        Subclasses that want to be notified when an interface is closed can
        declare a method named close_{name}_interface that will be called when
        the given interface is opened.  If such a function is declared and
        raises an exception, the interface will still be closed but the
        exception will be raised to the caller.

        Args:
            name (str): The name of the interface to close.
        """

        if name not in self._interface_status:
            raise ArgumentError("Unknown interface name in close_interface: {}".format(name))

        iface_close_name = "_close_{}_interface".format(name)

        self._interface_status[name] = False

        if hasattr(self, iface_close_name):
            close_func = getattr(self, iface_close_name)
            res = close_func()
            if inspect.isawaitable(res):
                await res

    def push_script_chunk(self, chunk):
        """Called when someone pushes a new bit of a TRUB script to this device

        Args:
            chunk (str): a buffer with the next bit of script to append
        """

    async def async_rpc(self, address, rpc_id, payload=b""):
        """Call an RPC by its address and ID.

        Subclasses must override this function with methods of finding RPCs as
        part of their implementations.  This method is designed to be called
        from a DeviceAdapter subclass and passed binary encoded arguments.

        It is not intended to be called directly by users.  If you are looking
        for a simple way to test out the rpcs provided by a VirtualDevice
        without needing an event loop or worrying about packing, you should
        use :meth:`simple_rpc`.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            bytes: The response payload from the RPC
        """

        raise RPCNotFoundError("RPC not found because virtual device did not implement call_rpc method")

    def simple_rpc(self, address, rpc_id, *args, **kwargs):
        """Synchronously dispatch an RPC inside this device.

        This function is meant to be used for testing purposes.  It is not
        an async method and indeed cannot be called from an event loop since
        it synchronously blocks until the RPC finished before returning.

        Callers of this function must know the argument spec of the RPC that
        they are trying to call and expected return value.

        Args:
            address (int): The address of the tile that has the RPC.
            rpc_id (int or RPCDeclaration): The 16-bit id of the rpc we want
                to call. If an RPCDeclaration is passed then arg_format and
                result_format are not required since they are included in the
                RPCDeclaration.
            *args: Any required arguments for the RPC as python objects.
            **kwargs: Only two keyword arguments are supported:
                - arg_format: A format specifier for the argument list
                - result_format: A format specifier for the result

        Returns:
            list: A list of the decoded response members from the RPC.
        """

        if isinstance(rpc_id, RPCDeclaration):
            arg_format = rpc_id.arg_format
            resp_format = rpc_id.resp_format
            rpc_id = rpc_id.rpc_id
        else:
            arg_format = kwargs.get('arg_format', None)
            resp_format = kwargs.get('resp_format', None)

        arg_payload = b''

        if arg_format is not None:
            arg_payload = pack_rpc_payload(arg_format, args)

        self._logger.debug("Sending rpc to %d:%04X, payload=%s", address, rpc_id, args)

        resp_payload = self._loop.run_coroutine(self.async_rpc(address, rpc_id, arg_payload))
        if resp_format is None:
            return []

        return unpack_rpc_payload(resp_format, resp_payload)



"""Base class for objects that have a background runnable task like virtual devices and tiles."""

import asyncio
import logging
import inspect
from iotile.core.exceptions import InternalError
from iotile.core.utilities.async_tools import SharedLoop, BackgroundTask


class _PeriodicWorkerMixin:
    """A simple mixin class that can run periodic callbacks in coroutines.

    This is not designed to be a general purpose class and requires that
    the class it is mixed into have a _loop and _started property.
    """

    def __init__(self):
        self._queued_workers = []
        self._workers = []

    def create_worker(self, func, interval, *args, **kwargs):
        """Spawn a worker thread running func.

        The worker will be automatically be started when start() is called and
        terminated when stop() is called on this object. This must be called
        only from the main thread.

        ``create_worker`` must not be called after stop() has been called.  If
        it is called before start() is called, the worker is started when
        start() is called, otherwise it is started immediately.

        Args:
            func (callable): Either a function that will be called in a loop
                with a sleep of interval seconds with *args and **kwargs or
                a generator function that will be called once and expected to
                yield periodically so that the worker can check if it should
                be killed.  If the function returns an awaitable, it will be
                awaited before sleeping again
            interval (float): The time interval between invocations of func.
                This should not be 0 so that the thread doesn't peg the CPU
                and should be short enough so that the worker checks if it
                should be killed in a timely fashion.
            *args: Arguments that are passed to func as positional args
            **kwargs: Arguments that are passed to func as keyword args
        """

        worker = (func, interval, args, kwargs)

        if not self._started:
            self._queued_workers.append(worker)
            return

        self._start_worker(worker)

    def _start_worker(self, worker):
        func, interval, args, kwargs = worker
        task = BackgroundTask(_worker_main(func, interval, args, kwargs), loop=self._loop)
        self._workers.append(task)

    def start_workers(self):
        """Start running this virtual device including any necessary worker threads."""

        for worker in self._queued_workers:
            self._start_worker(worker)

    def stop_workers(self):
        """Synchronously stop any potential workers."""

        for worker in self._workers:
            worker.stop_threadsafe()


async def _worker_main(func, interval, args, kwargs):
    logger = logging.getLogger(__name__)

    try:
        while True:
            await asyncio.sleep(interval)

            try:
                res = func(*args, **kwargs)
                if inspect.isawaitable(res):
                    await res
            except asyncio.CancelledError:
                raise
            except:  #pylint:disable=bare-except; We want to capture all errors in the background worker
                logger.exception("Error running background worker %s with args=%s, kwargs=%s", func, args, kwargs)
    except asyncio.CancelledError:
        pass
    except:
        logger.exception("Unknown exception in background worker")
        raise
